"""
federation/rate_limit.py
Federation rate limiting implementation.

Features:
- Per-domain rate limiting
- Multiple rate limit strategies
- In-memory storage
- Configurable limits
- Burst handling
"""

from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from enum import Enum
from http import HTTPStatus
from logging import Logger
from urllib.parse import urlparse

from fastapi.responses import JSONResponse

from phederation.cache import BaseCache
from phederation.utils.base import ObjectId, validate_object_id, RateLimit
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings
from phederation.utils.version import PHEDERATION_HEADERS


class RateLimitStrategy(Enum):
    """Rate limit strategies."""

    FIXED_WINDOW = "fixed_window"
    SLIDING_WINDOW = "sliding_window"
    TOKEN_BUCKET = "token_bucket"
    LEAKY_BUCKET = "leaky_bucket"


@dataclass
class RateLimitState:
    """Current rate limit state for a domain.

    Counts how many requests are `remaining` until a certain `reset` date.
    """

    remaining: int
    """ Number of remaining requests during a set period. """

    reset: datetime
    """ The time to reset the remaining requests counter. """

    limit: int
    """ The maximum for this rate limit state, to reset to at `reset` time. Set from the general `RateLimit`. """


class RateLimiter:
    """Federation rate limiting."""

    EVERYBODY: ObjectId = validate_object_id(url="http://everybody.localhost")
    """ The generic id to identify a non-identified domain. This rate limits "everybody", i.e. sets the limit on requests on non-identified urls. """

    def __init__(
        self,
        cache: BaseCache,
        settings: PhedSettings,
        strip_domains: bool = True,
    ):
        """Initialize rate limiter.

        Args:
            cache: The cache to use for storing urls
            default_limit: Default rate limit settings
            strategy: Rate limiting strategy
            strip_domains: If input domains should be stripped before caching. This can be used to either rate limit entire domains or specific actors.
        """
        self.settings: PhedSettings = settings
        # TODO: implement rate limiting strategy and load from settings
        self.strategy: RateLimitStrategy = RateLimitStrategy.FIXED_WINDOW
        self.policies: dict[str, RateLimit] = self.settings.rate_limiter.policies
        self.default_policy: RateLimit = self.settings.rate_limiter.default_policy
        self.cache: BaseCache = cache
        self.strip_domains: bool = True
        self.logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix)

    def _strip_domain(self, domain: str):
        domain = domain if "://" in domain else f"https://{domain}"
        if self.strip_domains:
            parsed_domain = urlparse(domain)
            return parsed_domain.netloc
        else:
            return domain

    def _get_policy(self, domain: str) -> RateLimit:
        if policy := self.policies.get(domain, None):
            return policy
        return self.default_policy

    def _get_state(self, domain: str) -> RateLimitState:
        """Get current rate limit state for domain."""
        domain = self._strip_domain(domain)
        state = self.cache.get(f"ratelimit:{domain}")
        if not state:
            policy = self._get_policy(domain)
            state = RateLimitState(
                remaining=policy.requests,
                reset=datetime.now(timezone.utc) + timedelta(seconds=policy.period),
                limit=policy.requests,
            )
            self._set_state(domain, state)
        return state

    def _set_state(self, domain: str, state: RateLimitState) -> None:
        """Set rate limit state for domain."""
        domain = self._strip_domain(domain)
        self.cache.set(f"ratelimit:{domain}", state)

    def ratelimit_response(self, domain: str) -> JSONResponse:
        headers = self.update_headers(domain=domain, headers=PHEDERATION_HEADERS)
        return JSONResponse(headers=headers, content={"status": "error", "detail": "rate limit reached"}, status_code=HTTPStatus.TOO_MANY_REQUESTS)

    def check_rate_limit(self, domain: str, reduce_counter: bool = True, strip_domain: bool = True) -> bool:
        """Check limit for given domain, and reduce request counter by one."""
        self.strip_domains = strip_domain
        domain = self._strip_domain(domain)
        state = self._get_state(domain)

        # Reset if expired
        if datetime.now(timezone.utc) >= state.reset:
            policy = self._get_policy(domain)
            state = RateLimitState(
                remaining=policy.requests,
                reset=datetime.now(timezone.utc) + timedelta(seconds=policy.period),
                limit=policy.requests,
            )
            self._set_state(domain, state)
            self.strip_domains = True
            return True

        if state.remaining <= 0:
            self.strip_domains = True
            return False

        # Update remaining
        if reduce_counter:
            state.remaining -= 1
            self._set_state(domain, state)
        self.strip_domains = True
        return True

    def get_wait_time(self, domain: str) -> float:
        """Get wait time in seconds until rate limit reset."""
        domain = self._strip_domain(domain)
        state = self._get_state(domain)
        if state.remaining > 0:
            return 0

        wait_time = (state.reset - datetime.now(timezone.utc)).total_seconds()
        return max(0, wait_time)

    def update_rate_limit(self, domain: str, headers: dict[str, str]) -> RateLimitState | None:
        """Update rate limit state from response headers.

        This allows remote instances to update the rate limits of themselves.
        Will be limited below by one to always be able to send at least one request.

        Header keys/values
         - X-RateLimit-Remaining (int): number of remaining requests that the remote instance accepts.
         - X-RateLimit-Reset (int): the reset timestamp when the remote request counter resets.
         - X-RateLimit-Limit (int): the total limit of requests accepted by the remote instance until the reset date.
        """
        domain = self._strip_domain(domain)
        headers_lower = {key.lower(): headers[key] for key in headers.keys()}
        remaining = headers_lower.get("x-ratelimit-remaining")
        reset = headers_lower.get("x-ratelimit-reset")
        limit = headers_lower.get("x-ratelimit-limit")

        if remaining is not None and reset is not None and limit is not None:
            state = RateLimitState(remaining=int(remaining), reset=datetime.fromtimestamp(float(reset), tz=timezone.utc), limit=max(1, int(limit)))
            self._set_state(domain, state)
            return state

    def update_headers(self, domain: str, headers: dict[str, str]):
        domain = self._strip_domain(domain)
        state = self._get_state(domain=domain)
        headers["x-ratelimit-remaining"] = str(state.remaining)
        headers["x-ratelimit-reset"] = str(state.reset.timestamp())
        headers["x-ratelimit-limit"] = str(state.limit)
        if state.remaining == 0:
            headers["retry-after"] = str(int((state.reset - datetime.now(timezone.utc)).total_seconds() + 1))
        return headers

    def clear(self, domain: str) -> None:
        domain = self._strip_domain(domain)
        """Clear rate limit state for domain."""
        self.cache.delete(f"ratelimit:{domain}")



import asyncio
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from http import HTTPStatus
from typing import Any, cast, override

from temporalio import activity, workflow
from temporalio.client import (
    Schedule,
    ScheduleActionStartWorkflow,
    ScheduleHandle,
    ScheduleIntervalSpec,
    ScheduleSpec,
    ScheduleState,
)
from temporalio.common import RetryPolicy
from temporalio.exceptions import ApplicationError

from phederation.utils.exceptions import DeliveryError

with workflow.unsafe.imports_passed_through():

    from phederation.utils.base import AccessType
    from phederation.utils.base import UrlType
    from phederation.utils.serialization import to_json_string
    from phederation.utils.logging import configure_logger
    from phederation.models.activities import APAnnounce
    from phederation.utils.base import ObjectId
    from phederation.utils.exceptions import ActivityPubException
    from phederation.federation.tasks import PeriodicBackgroundTask
    from phederation.models import dereference
    from phederation.federation.delivery import ActivityDelivery, DeliveryResult
    from phederation.models import ValidCollection
    from phederation.models.activities import APActivity
    from phederation.models.collections import APOrderedCollection
    from phederation.utils.base import ObjectId, collection_id_from_name
    from phederation.utils.exceptions import (
        DiscoveryError,
        ResolverError,
        UserError,
    )
    from phederation.utils.settings import FederationSettings


@workflow.defn
class UpdateNodeInfoWorkflow:
    """Workflow to periodically update the NodeInfo of a running instance.

    It is usually used in a temporalio scheduler, and started by the `UpdateNodeInfoTask`.
    """

    @workflow.run
    async def run(
        self,
    ) -> ObjectId:
        """Run the workflow.

        Returns:
            ObjectId: The id of the NodeInfo document.
        """
        retry_policy = RetryPolicy(initial_interval=timedelta(seconds=1), maximum_attempts=3)
        document_id = await workflow.execute_activity_method(
            UpdateNodeInfoTask.update_node_info,
            retry_policy=retry_policy,
            schedule_to_close_timeout=timedelta(seconds=20),
        )
        print(f"UpdateNodeInfoWorkflow: Updated instance node information.")
        return document_id


@workflow.defn
class KeyRotationWorkflow:
    """Workflow to periodically rotate keys of a running instance.

    It is usually used in a temporalio scheduler, and started by the `KeyRotationTask`.
    """

    @workflow.run
    async def run(self, federation_settings: FederationSettings):
        """Run the workflow."""
        retry_policy_rotate = RetryPolicy(initial_interval=timedelta(seconds=10), maximum_attempts=1)
        retry_policy_announce = RetryPolicy(
            initial_interval=timedelta(seconds=5), backoff_coefficient=2.0, maximum_attempts=federation_settings.max_retries
        )

        logger = configure_logger(name=__name__, prefix=federation_settings.logging_prefix)

        logger.debug("Rotating keys")
        key_rotation_parameters = await workflow.execute_activity_method(
            KeyRotationTask.rotate_keys,
            retry_policy=retry_policy_rotate,
            schedule_to_close_timeout=timedelta(seconds=30),
        )
        logger.debug(f"Announcing rotated keys: {key_rotation_parameters}")
        for params in key_rotation_parameters:
            print(params)
            _ = await workflow.execute_activity_method(
                KeyRotationTask.announce_key_rotation,
                params,
                retry_policy=retry_policy_announce,
                schedule_to_close_timeout=timedelta(seconds=30),
            )
        logger.info(f"KeyRotationWorkflow: Rotated instance keys.")


@dataclass
class PrepareWorkflowParams:
    settings_federation: FederationSettings
    activity: APActivity
    recipients: list[ObjectId]


@dataclass
class DeliveryWorkflowParams:
    activity: APActivity
    inbox: ObjectId


class DeliveryWorkerTask(PeriodicBackgroundTask):
    def __init__(self, update_period: int):
        super().__init__(update_period)
        self.schedule_handle: ScheduleHandle | None = None

    @override
    async def on_startup(self) -> None:
        if not self.server:
            self.logger.warning("Trying to start DeliveryWorkerTask without initialized server.")
            return

        if not self.server.settings.worker.enabled or not self.server.worker_client:
            self.logger.warning(
                "Trying to start DeliveryWorkerTask without enabling temporalio worker. Schedule will not be started, using internal periodic task."
            )
            return

        self.schedule_handle = await self.server.worker_client.create_schedule(
            "workflow-schedule-DeliveryWorkerTask",
            Schedule(
                action=ScheduleActionStartWorkflow(
                    ActivityDeliveryWorkflow.run,
                    self.server.settings.federation,
                    id="schedules-workflow-DeliveryWorkerTask",
                    task_queue=self.server.settings.worker.tasks_queue_name,
                ),
                spec=ScheduleSpec(
                    intervals=[ScheduleIntervalSpec(every=timedelta(seconds=self.server.settings.federation.delivery_queue_polling_period))]
                ),
                state=ScheduleState(note="Periodic polling of delivery queue for new messages to send."),
            ),
        )

        self.logger.info(f"Scheduled updating node info data every {self.server.settings.federation.update_nodeinfo_period} seconds.")

    @override
    async def on_update(self) -> None:
        """Background task for polling messages to send, if no temporalio schedule was created."""
        await super().on_update()
        if self.schedule_handle:
            # if there is a temporalio schedule do not update here
            return

        params = await self.poll_delivery_queue(self.server.settings.federation)
        if not params:
            # self.logger.info(f"DeliveryWorkerTask: Polled messages, nothing to do.")
            return

        inboxes = await self.prepare_inboxes_for_delivery(params)
        results = await asyncio.gather(*[self.deliver_activity(DeliveryWorkflowParams(activity=params.activity, inbox=inbox)) for inbox in inboxes])

        self.logger.info(f"DeliveryWorkerTask: Polled messages, result status_code {DeliveryResult.create_from_list(results).status_code}.")

    @override
    async def on_shutdown(self):
        await super().on_shutdown()
        if self.schedule_handle:
            await self.schedule_handle.delete()
            self.schedule_handle = None

    @activity.defn
    async def poll_delivery_queue(self, settings_federation: FederationSettings) -> None | PrepareWorkflowParams:
        queue = self.server.delivery.delivery_queue
        if not queue:
            raise DeliveryError("Server delivery does not have a delivery queue set up to poll from")

        activity_in_queue = await queue.pop()
        if not activity_in_queue:
            return None
        activity = activity_in_queue.activity
        recipients = activity_in_queue.recipients

        return PrepareWorkflowParams(settings_federation=settings_federation, activity=activity, recipients=recipients)

    @activity.defn
    async def prepare_inboxes_for_delivery(self, params: PrepareWorkflowParams) -> list[ObjectId]:
        # deduplicate recipient list before resolving their inboxes
        recipients = list(set(params.recipients))

        # loop over recipients, carefully checking if they have an inbox
        recipient_inboxes: list[ObjectId] = []
        for recipient in recipients:
            try:
                inbox = await self.server.resolver.resolve_inboxes([recipient])
                recipient_inboxes.extend(inbox)
            except UserError as e:
                self.logger.error(f"Could not resolve inbox for recipient {recipient}: {e.message}")
            except ResolverError as e:
                self.logger.error(f"Could not resolve inbox for recipient {recipient}: {e.message}")

        # Servers MUST de-duplicate the final recipient list.
        # https://www.w3.org/TR/activitypub/#delivery
        recipient_inboxes = list(set(recipient_inboxes))

        return recipient_inboxes

    @activity.defn
    async def deliver_activity(self, params: DeliveryWorkflowParams) -> DeliveryResult:
        result = await self.server.delivery.deliver_to_url(activity=params.activity, url=params.inbox, retry_count=0)
        if result.success:
            return result
        else:
            raise ApplicationError(f"Delivery failed for inbox {params.inbox}. Result error: {result.error_message}")


@workflow.defn
class ActivityDeliveryWorkflow:

    @workflow.run
    async def run(self, settings_federation: FederationSettings):
        logger = configure_logger(name=__name__, prefix=settings_federation.logging_prefix)
        logger.debug(f"ActivityDeliveryWorkflow: Polling delivery queue...")

        retry_policy = RetryPolicy(initial_interval=timedelta(seconds=3), maximum_attempts=1)
        params = await workflow.execute_activity_method(
            DeliveryWorkerTask.poll_delivery_queue,
            settings_federation,
            retry_policy=retry_policy,
            schedule_to_close_timeout=timedelta(seconds=settings_federation.delivery_timeout),
        )
        if not params:
            logger.debug(f"ActivityDeliveryWorkflow: Nothing to do.")
            return
        
        logger.info(f"ActivityDeliveryWorkflow: Starting delivery to {len(params.recipients)} recipients...")
        # first, make sure that all bcc and bto recipients are actually removed
        params.activity = await ActivityDelivery.remove_hidden_recipients(params.activity)

        # then prepare all inboxes for all recipients, deduplicating the list in the process
        retry_policy = RetryPolicy(initial_interval=timedelta(seconds=3), maximum_attempts=params.settings_federation.max_retries)
        recipient_inboxes = await workflow.execute_activity_method(
            DeliveryWorkerTask.prepare_inboxes_for_delivery,
            params,
            retry_policy=retry_policy,
            schedule_to_close_timeout=timedelta(seconds=params.settings_federation.delivery_timeout),
        )
        logger.debug(f"Prepared inboxes of {len(recipient_inboxes)} recipients")

        # finally, send off the activity to each recipient
        retry_policy = RetryPolicy(
            initial_interval=timedelta(seconds=1), backoff_coefficient=2.0, maximum_attempts=params.settings_federation.max_retries
        )
        handles: list[workflow.ActivityHandle[DeliveryResult]] = []
        for inbox in recipient_inboxes:
            logger.debug(f"Sending activity to {inbox}...")
            handle = workflow.start_activity_method(
                DeliveryWorkerTask.deliver_activity,
                DeliveryWorkflowParams(activity=params.activity, inbox=inbox),
                retry_policy=retry_policy,
                schedule_to_close_timeout=timedelta(seconds=params.settings_federation.delivery_timeout),
                summary="Deliver a given activity to a given inbox",
            )
            handles.append(handle)
        _ = await asyncio.gather(*handles)

        # wait on all send activities to be done
        await workflow.wait_condition(lambda: workflow.all_handlers_finished())

        logger.debug(f"Every message sent, collecting results")
        # collect all results
        delivery_results: list[DeliveryResult] = []
        for h in handles:
            if h.done() and not h.cancelled():
                delivery_results.append(h.result())
            else:
                # context = h.get_context()
                if params.activity.id:
                    delivery_results.append(DeliveryResult(failed=[params.activity.id], status_code=HTTPStatus.INTERNAL_SERVER_ERROR))

        logger.info(f"ActivityDeliveryWorkflow: Delivery done.")
        return delivery_results


class UpdateNodeInfoTask(PeriodicBackgroundTask):

    def __init__(self, update_period: int):
        super().__init__(update_period)
        self.schedule_handle: ScheduleHandle | None = None

    @override
    async def on_startup(self) -> None:
        if not self.server:
            self.logger.warning("Trying to update node info task without initialized task.")
            return

        if not self.server.settings.worker.enabled or not self.server.worker_client:
            self.logger.warning("Trying to update node info task without enabling temporalio worker, using internal periodic task.")
            return

        self.schedule_handle = await self.server.worker_client.create_schedule(
            "workflow-schedule-UpdateNodeInfoTask",
            Schedule(
                action=ScheduleActionStartWorkflow(
                    UpdateNodeInfoWorkflow.run,
                    id="schedules-workflow-UpdateNodeInfoTask",
                    task_queue=self.server.settings.worker.tasks_queue_name,
                ),
                spec=ScheduleSpec(intervals=[ScheduleIntervalSpec(every=timedelta(seconds=self.server.settings.federation.update_nodeinfo_period))]),
                state=ScheduleState(note="Periodic update of the instance node info data."),
            ),
        )

        self.logger.info(f"Scheduled updating node info data every {self.server.settings.federation.update_nodeinfo_period} seconds.")

    @override
    async def on_update(self) -> None:
        """Background task for updating node info, if no temporalio schedule was created."""
        await super().on_update()
        if self.schedule_handle:
            # if there is a temporalio schedule, do not update here
            return

        node_info_id = await self.update_node_info()

        self.logger.info(f"UpdateNodeInfoTask: Updated node info, new id {node_info_id}")

    @override
    async def on_shutdown(self):
        await super().on_shutdown()
        if self.schedule_handle:
            await self.schedule_handle.delete()
            self.schedule_handle = None

    @activity.defn
    async def update_node_info(
        self,
    ) -> ObjectId:
        try:
            self.logger.debug("Updating node info...")
            if not self.server:
                raise DiscoveryError("Server of update info task not setup")
            settings = self.server.settings
            discovery = self.server.discovery

            collection_id = collection_id_from_name(id=settings.domain.base_url, name=UrlType.Users.value, page=None)

            node_info = await discovery.discover_nodeinfo(base_url=settings.domain.base_url)
            if node_info is None:
                raise DiscoveryError("Could not obtain NodeInfo from local instance")
            users_response = await discovery.requests.get(url=collection_id, headers=discovery.headers)
            users_response = users_response.raise_for_status()

            users_dict: dict[str, Any] = cast(dict[str, Any], users_response.json())
            users_collection = APOrderedCollection.deserialize(users_dict)
            if not isinstance(users_collection, APOrderedCollection):
                raise DiscoveryError(f"Users collection could not be retrieved properly (was type {type(users_collection)})")
            node_info.usage.users.total = users_collection.total_items or 0
            node_info.updated_at = datetime.now(timezone.utc)
            node_info.id = None # so that a new node info object gets created

            document_id = await self.server.storage.node_info.create(data=node_info)
        except ActivityPubException as e:
            raise ApplicationError(e.message)
        return document_id


@dataclass
class KeyRotationAnnounceParams:
    # The id of the new key.
    key_id_old: ObjectId
    # The id of the old key, used to sign the request for update.
    key_id_new: ObjectId
    # The actor associated to the key.
    actor_id: ObjectId


class KeyRotationTask(PeriodicBackgroundTask):

    def __init__(self, enabled: bool, update_period: int):
        super().__init__(update_period)
        self.enabled: bool = enabled
        self.schedule_handle: ScheduleHandle | None = None

    @override
    async def on_startup(self):
        await super().on_startup()

        if not self.enabled:
            self.logger.info("Key rotation task disabled.")
            return

        if not self.server:
            self.logger.warning("Trying to start key rotation task without server.")
            return

        if not self.server.settings.worker.enabled or not self.server.worker_client:
            self.logger.warning("Trying to run key rotation task without enabling temporalio worker. Schedule will not be started.")
            return

        self.schedule_handle = await self.server.worker_client.create_schedule(
            "workflow-schedule-KeyRotationTask",
            Schedule(
                action=ScheduleActionStartWorkflow(
                    KeyRotationWorkflow.run,
                    self.server.settings.federation,
                    id="schedules-workflow-KeyRotationTask",
                    task_queue=self.server.settings.worker.tasks_queue_name,
                ),
                spec=ScheduleSpec(intervals=[ScheduleIntervalSpec(every=timedelta(seconds=self.server.settings.security.key_rotation_period))]),
                state=ScheduleState(note="Periodic update of the instance key data."),
            ),
        )

        self.logger.info(f"Scheduled rotation of actor keys every {self.server.settings.security.key_rotation_period} seconds.")

    @override
    async def on_update(self) -> None:
        """Background task for key rotation, if no temporalio schedule was created."""
        if self.schedule_handle or not self.enabled:
            # if there is a temporalio schedule, or key rotation is not enabled, do not update here
            return

        await super().on_update()
        key_rotation_parameters = await self.rotate_keys()
        for params in key_rotation_parameters:
            await self.announce_key_rotation(params)
        self.logger.info(f"KeyRotationTask: Rotated instance keys.")

    @override
    async def on_shutdown(self):
        await super().on_shutdown()
        if self.server:
            await self.server.key_manager.close()
        if self.schedule_handle:
            await self.schedule_handle.delete()

    @activity.defn
    async def announce_key_rotation(self, args: KeyRotationAnnounceParams) -> None:
        """Announce new key to federation.

        This creates an APUpdate activity with the actor set to the old keys actor id,
        the new key APPublicKey object as the "object", and signs the activity proof and HTTP signature with the OLD
        key so that servers receiving it can verify that was indeed the actor that signed it.

        Args:
            args (KeyRotationAnnounceParams): Parameters for key rotation announcement.
        """
        try:
            self.logger.debug(f"Announcing key {args.key_id_new}")
            # sign the update with the old key, KEEP THE OLD KEY for a while (two months? longer?) so that the federated servers can verify the update
            key_new = await self.server.key_manager.get_public_key_APObject(key_id=args.key_id_new)

            string_to_sign = to_json_string(key_new.serialize())
            assert string_to_sign, "Could not properly serialize the new key"

            proof = await self.server.key_manager.proof_string(string_to_sign, private_key_id=args.key_id_old)

            # Announce activity, with the proof added to it.
            # The object is the new key, the actor sending it is the instance actor. It it sent to the actor on the local instance,
            # and bto will be "all federated instances", set in the deliver_to_federation method.
            announce_new_key = APAnnounce(
                actor=self.server.instance.actor_id,
                to=[args.actor_id],
                object=key_new,
                summary="Key rotation activity. Contains the new key to announce, and a proof of the serialized key object, signed with the old key.",
                proof=proof,
            )
            await self.server.delivery.deliver_to_federation(activity=announce_new_key)
            self.logger.debug(f"Completed announcing key {args.key_id_new} to federation")
        except ActivityPubException as e:
            # use the actual error message instead of user facing here, to be able to debug temporalio workflow better
            raise ApplicationError(e.message)

    @activity.defn
    async def rotate_keys(
        self,
    ) -> list[KeyRotationAnnounceParams]:
        """Checks key expiry and rotates expired keys.
        This generates a new key pair, removes expired keys, and returns parameters to announce new keys to federation in the next workflow step.

        Raises:
            KeyError: If the server of the task is not set up.

        Returns:
            list[KeyRotationAnnounceParams]: List of parameters to announce newly generated keys to the federation.
        """
        try:
            self.logger.debug("Rotating instance keys...")
            if not self.server:
                raise KeyError("Server of key rotation task not setup")

            # iterate over the pages of all actors, replace their keys if necessary
            collection_id = collection_id_from_name(id=self.server.base_url, name=ValidCollection.Users.value)
            rotation_results: list[KeyRotationAnnounceParams] = []
            async for page in self.server.collections.pages(collection_id=collection_id, access=AccessType.PRIVATE):
                # get all item ids of all keys on the page
                actor_ids = page.ordered_items or page.items or []
                actor_ids = [dereference(k, "id") for k in actor_ids if k]
                actor_ids = [k for k in actor_ids if k]

                # replace expired keys locally, asynchronously
                async def _rotate_key_of_actor(actor_id: ObjectId) -> KeyRotationAnnounceParams | None:
                    key_id = await self.server.key_manager.get_key_id_from_actor(actor_id)
                    try:
                        key = await self.server.key_manager.resolve_key(key_id=key_id)
                        if key.is_expired(datetime.now(timezone.utc)):
                            self.logger.debug(f"Rotating instance key of actor {actor_id}")
                            key_id_old = key.id
                            assert key_id_old, "key_id_old is None"
                            key_id_new = await self.server.key_manager.rotate_key(key=key, actor_id=actor_id)
                            return KeyRotationAnnounceParams(key_id_new=key_id_new, key_id_old=key_id_old, actor_id=actor_id)
                    except ActivityPubException as e:
                        self.logger.warning(f"Could not resolve key from actor {actor_id}, error {e.message}")
                    except Exception as e:
                        self.logger.warning(f"Could not resolve key from actor {actor_id}, error {e}")

                for actor_id in actor_ids:
                    try:
                        result = await _rotate_key_of_actor(actor_id)
                        if result:
                            rotation_results.append(result)
                    except ActivityPubException as e:
                        self.logger.warning(e.message)

            # TODO: remove old keys that are expired for a long time
            # TODO: make this activity "per page", so that several workers can work on different pages

            # clear key cache
            if self.server.key_manager.cache:
                self.server.key_manager.cache.clear()
            return rotation_results
        except ActivityPubException as e:
            # use the actual error message instead of user facing here, to be able to debug temporalio workflow better
            raise ApplicationError(e.message)



from collections.abc import Awaitable, Callable
from http import HTTPStatus
from typing import override

from .base import ResolverResult


class ActivityPubException(Exception):
    """Base exception for ActivityPub-related errors."""

    default_message: str = "Internal server error"
    message: str
    user_message: str | None = None
    status_code: HTTPStatus = HTTPStatus.INTERNAL_SERVER_ERROR

    def __init__(self, message: str, user_facing_message: str | None = None, status_code: HTTPStatus | None = None, *args: object) -> None:
        self.message = message
        if not user_facing_message:
            user_facing_message = self.default_message
        self.user_message = user_facing_message
        if status_code:
            self.status_code = status_code
        super().__init__(user_facing_message, *args)

    @override
    def __str__(self) -> str:
        return self.user_message or "Unknown APException"


def catch_exceptions(exception_class: type[ActivityPubException], message: str):
    def with_exception_handled[**Params, ReturnType](function: Callable[Params, Awaitable[ReturnType]]) -> Callable[Params, Awaitable[ReturnType]]:
        async def wrapper_method(*args_with_exception: Params.args, **kwargs_with_exception: Params.kwargs) -> ReturnType:
            try:
                return await function(*args_with_exception, **kwargs_with_exception)
            except ActivityPubException as e:
                e.message = f"{message}; args={args_with_exception}. {e.message}"
                e.user_message = message
                raise e
            except Exception as e:
                raise exception_class(
                    f"Exception {type(e).__name__}: {message}. Args: {args_with_exception}",
                    user_facing_message=message,
                )

        return wrapper_method

    return with_exception_handled


class InternalLoggingException(ActivityPubException):
    """An exception that should not be transmitted over the network.
    The message is not 'safe' to transmit and show publicly (may contain sensitive data)."""


class ValidationError(InternalLoggingException):
    """Raised when validation fails."""

    default_message: str = "Validation failed"
    status_code: HTTPStatus = HTTPStatus.BAD_REQUEST


class InvalidURLError(ValidationError):
    """Raised when an invalid URL is provided."""

    default_message: str = "Wrong URL"
    status_code: HTTPStatus = HTTPStatus.BAD_GATEWAY


class DecodingError(ValidationError):
    """Raised when a string cannot be decoded to JSON or from bytes."""

    default_message: str = "Cannot decode"
    status_code: HTTPStatus = HTTPStatus.BAD_REQUEST


class SignatureError(InternalLoggingException):
    """Raised when signature creation fails."""

    default_message: str = "Wrong signature"
    status_code: HTTPStatus = HTTPStatus.UNAUTHORIZED


class AuthenticationError(InternalLoggingException):
    """Raised when authentication fails."""

    default_message: str = "Not authenticated"
    status_code: HTTPStatus = HTTPStatus.UNAUTHORIZED


class AuthorizationError(InternalLoggingException):
    """Raised when authorization fails."""

    default_message: str = "Not authorized"
    status_code: HTTPStatus = HTTPStatus.UNAUTHORIZED


class RateLimitExceeded(InternalLoggingException):
    """Raised when rate limit is exceeded."""

    default_message: str = "Rate limit exceeded"
    status_code: HTTPStatus = HTTPStatus.TOO_MANY_REQUESTS


class SecurityError(InternalLoggingException):
    """Raised when security-related errors occur."""

    default_message: str = "Security issue"
    status_code: HTTPStatus = HTTPStatus.UNAUTHORIZED


class HandlerError(InternalLoggingException):
    """Raised when handling fails"""

    default_message: str = "Failed to handle activity"
    status_code: HTTPStatus = HTTPStatus.BAD_REQUEST


class StorageError(InternalLoggingException):
    """Raised when storage-related errors occur."""

    default_message: str = "Failed to load"
    status_code: HTTPStatus = HTTPStatus.INTERNAL_SERVER_ERROR


class DiscoveryError(InternalLoggingException):
    """Raised when discovery fails."""

    default_message: str = "Discovery failed"
    status_code: HTTPStatus = HTTPStatus.BAD_REQUEST


class ResolverError(InternalLoggingException):
    """Raised when resolver-related errors occur."""

    default_message: str = "Failed to resolve"
    status_code: HTTPStatus = HTTPStatus.BAD_REQUEST
    response: ResolverResult | None = None

    def __init__(
        self,
        message: str,
        user_facing_message: str | None = None,
        status_code: HTTPStatus | None = None,
        response: ResolverResult | None = None,
        *args: object,
    ) -> None:
        self.response = response
        super().__init__(message, user_facing_message, status_code, *args)

    def get(self, key: str):
        if self.response:
            return self.response.get(key)


class DeliveryError(InternalLoggingException):
    """Raised when delivery-related errors occur."""

    default_message: str = "Failed to deliver"
    status_code: HTTPStatus = HTTPStatus.BAD_REQUEST


class KeyManagementError(InternalLoggingException):
    """Raised when key manager-related errors occur."""

    default_message: str = "Key management error"
    status_code: HTTPStatus = HTTPStatus.BAD_REQUEST


class CollectionError(InternalLoggingException):
    """Raised when collection-related errors occur."""

    default_message: str = "Collection error"
    status_code: HTTPStatus = HTTPStatus.INTERNAL_SERVER_ERROR


class IntegrationError(InternalLoggingException):
    """Raised when integration-related errors occur."""

    default_message: str = "Integration error"
    status_code: HTTPStatus = HTTPStatus.INTERNAL_SERVER_ERROR


class MiddlewareError(InternalLoggingException):
    """Raised when middleware-related errors occur."""

    default_message: str = "Middleware failed"
    status_code: HTTPStatus = HTTPStatus.BAD_GATEWAY


class UserError(InternalLoggingException):
    """Raised when a user related error occured."""

    default_message: str = "User not found"
    status_code: HTTPStatus = HTTPStatus.GONE


class APIError(InternalLoggingException):
    """Raised by an API call error."""

    default_message: str = "Failed to resolve API"
    status_code: HTTPStatus = HTTPStatus.BAD_GATEWAY


class MediaError(InternalLoggingException):
    """Raised by a media processing error."""

    default_message: str = "Failed to process media"
    status_code: HTTPStatus = HTTPStatus.BAD_REQUEST


class SettingsError(InternalLoggingException):
    """Raised by the instance if settings are poorly configured."""

    default_message: str = "Instance configuration failed"
    status_code: HTTPStatus = HTTPStatus.INTERNAL_SERVER_ERROR



"""
Enhanced Maintenance activity handler.
"""

from typing import cast, override

from phederation.handlers.create import CreateHandler
from phederation.models import dereference
from phederation.models.activities import APActivity
from phederation.models.maintenance import (
    APMaintenanceCommand,
    APMaintenanceMode,
    APMaintenanceModeCancel,
    MaintenanceCommandType,
)
from phederation.models.objects import APObject
from phederation.utils.base import AccessType, ActivityType, serialize_datetime
from phederation.utils.exceptions import HandlerError, ValidationError

from .base import ActivityHandler


class MaintenanceHandler(ActivityHandler):
    """Maintenance activity handler."""

    SUPPORTED_TYPES: dict[str, type[APMaintenanceMode] | type[APMaintenanceModeCancel]] = {
        MaintenanceCommandType.MAINTENANCE_MODE.value: APMaintenanceMode,
        MaintenanceCommandType.MAINTENANCE_MODE_CANCEL.value: APMaintenanceModeCancel,
    }

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """
        Maintenance command validation.

        Validates:
        - Activity structure
        - Object type support
        - Media attachments
        - Rate limits
        """
        # Validate activity.object
        if activity.type != ActivityType.MAINTENANCE.value:
            raise ValidationError(f"APActivity is not APMaintenance (is type {activity.type})")

        obj_dict = activity.object
        if isinstance(obj_dict, APObject):
            obj_dict = obj_dict.serialize()
        if not isinstance(obj_dict, dict):
            self.logger.warning(f"Object in activity was not an APObject (instead: {type(obj_dict)}): activity {activity.serialize()}")
            raise ValidationError(f"Object in MAINTENANCE did not have proper information")
        obj_type = cast(str, obj_dict.get("type", ""))

        if not obj_type or obj_type not in self.SUPPORTED_TYPES:
            raise ValidationError(f"Unsupported object type: {obj_type}")

        # Validate object model
        model_class = self.SUPPORTED_TYPES[obj_type]
        obj = model_class.model_validate(obj_dict)

        # Check actor permissions
        actor_id = dereference(activity.actor, key="id")
        if not actor_id:
            raise ValidationError("Create: APMaintenance.actor.id is None")

        actor = await self.resolver.resolve_actor(actor_id=actor_id)
        if not actor:
            raise ValidationError("Actor not found")

        # Attribute the object to the actor of Create
        if obj.attributed_to and obj.attributed_to != actor_id:
            raise ValidationError(f"Object in Maintenance is attributed to the wrong actor: attributed_to={obj.attributed_to}, actor_id={actor.id}")
        obj.attributed_to = actor_id

        # Validate hash
        await CreateHandler.validate_content_hash(self, obj)

        # Validate content
        await self._validate_content(obj)

        # Visibility of maintenance activity and objects is always private
        maximim_security_visibility = AccessType.PRIVATE.value

        obj.visibility = maximim_security_visibility

        # Important for "store_in_collection" of server to create the object at this point, and not too late in the handle_outbox method
        activity_object = await CreateHandler.store_object_locally(handler=self, obj=obj, actor_id=actor_id)
        activity.object = activity_object

        # Set the visibility of the activity to be at most the one from the object
        activity.visibility = maximim_security_visibility

        return activity

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """
        Maintenance processing.

        Handles:
        - Object storage
        - Side effects
        - Notifications
        - Federation
        """
        if not self.storage.settings.federation.admin_mode:
            raise HandlerError(
                f"Received Maintenance activity {activity.id} in outbox, but the instance is not in admin mode. Discarding the activity"
            )

        self.logger.info(f"Received Maintenance activity {activity.id} in outbox. Processing...")

        # Remove all recipients from the activity. A maintenance activity should never be sent to somebody, and certainly not federate.
        activity.clear_recipients()

        # actor_id = dereference_or_raise(activity, key="actor")
        obj = await self.resolver.resolve_object(activity.object)

        if not isinstance(obj, APMaintenanceCommand):
            raise HandlerError(f"Cannot handle command of type {type(obj).__name__} in maintenance handler")

        # Execute the maintenance command
        await self.execute_command(obj)

        return activity

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        self.logger.warning(
            f"Received MAINTENANCE in inbox (activity id: {activity.id}). This should not happen, discarding it",
            extra={"activity_id": activity.id},
        )

        return None

    async def _validate_content(self, object: APMaintenanceCommand) -> None:
        # TODO: implement content validation
        object = object

    async def execute_command(self, object: APMaintenanceCommand):
        if not self.storage.settings.federation.admin_mode:
            raise HandlerError(f"Trying to executing command {type(object).__name__}, but not in admin mode")
        if not self.resolver.instance:
            raise HandlerError(f"Instance object in resolver not set up, cannot access maintenance commands collection")
        maintenance_commands_id = self.resolver.instance.maintenance_commands

        self.logger.critical(f"Executing command {type(object).__name__}")

        if isinstance(object, APMaintenanceMode) and object.id:
            await self.collections.add_to_collection(collection_id=maintenance_commands_id, items=object.id, access=AccessType.PRIVATE)
            self.logger.critical(
                f"Initialize maintenance mode. Start time {serialize_datetime(object.start_time)}, end time {serialize_datetime(object.end_time)}, command id {object.id}",
                extra={"x-maintenance-id": object.id},
            )
        if isinstance(object, APMaintenanceModeCancel) and object.id:
            await self.collections.remove_from_collection(
                collection_id=maintenance_commands_id, items=object.command_to_cancel, access=AccessType.PRIVATE
            )
            self.logger.critical(f"Cancelling maintenance mode command with id {object.command_to_cancel}")



from enum import Enum
from http import HTTPStatus
from logging import Logger

from fastapi import HTTPException, Request
from fastapi.security.oauth2 import SecurityScopes
from pydantic import SecretStr

from phederation.security.authentication import UserInfo
from phederation.utils import configure_logger
from phederation.utils.settings import PhedSettings

from .authentication import Authentication


class PermissionType(str, Enum):
    NONE = "None"
    USER = "User"
    ADMIN = "Admin"


class Authorization:
    """Class to check the `PermissionType` for a given actor.
    Directly uses the OAuth2 Depends scheme, i.e. checks the token of the logged in user.
    """

    def __init__(self, settings: PhedSettings, authentication: Authentication):
        self.authentication: Authentication = authentication
        self.logger: Logger = configure_logger(__name__, prefix=settings.federation.logging_prefix)

    async def __call__(self, request: Request, security_scopes: SecurityScopes) -> UserInfo:
        """
        Authorization endpoint that requires a valid user info for access.
        Verifies that the logged in user has the requested security scopes as Realm Role in keycloak.
        """
        token = await self.authentication.oauth2_scheme_code(request)
        if not token:
            raise HTTPException(
                status_code=HTTPStatus.UNAUTHORIZED,
                detail="Not authenticated",
                headers={"WWW-Authenticate": "Bearer"},
            )
        token_secret = SecretStr(token)
        return await self.verify_token(token=token_secret, security_scopes=security_scopes)

    async def verify_token(self, token: SecretStr, security_scopes: SecurityScopes):
        self.logger.debug(f"Calling authorization endpoint, security_scopes={security_scopes.scopes}, token={token}")
        user_info: UserInfo | None = await self.authentication.verify_token(token)
        if not user_info:
            self.logger.debug(f"Could not get user info. Unauthorized")
            raise HTTPException(
                status_code=HTTPStatus.UNAUTHORIZED,
                detail="Permission level too low",
                headers={"WWW-Authenticate": "Bearer"},
            )

        # Verify that the user is allowed to access the requested scopes
        if not self.by_scope(user_info, security_scopes.scopes):
            raise HTTPException(
                status_code=HTTPStatus.UNAUTHORIZED,
                detail="Permission level too low",
                headers={"WWW-Authenticate": "Bearer"},
            )

        return user_info

    def by_scope(self, user_info: UserInfo, scopes: list[str]) -> bool:
        if not user_info.groups:
            self.logger.warning(f"UserInfo {user_info.preferred_username} does not include 'groups', cannot check authorization.")
            return False
        groups = [group.lower() for group in user_info.groups]
        for scope in scopes:
            if not user_info.groups or not scope in groups:
                self.logger.warning(f"User {user_info.preferred_username} tried to access permission level {scope}, not authorized.")
                raise HTTPException(
                    status_code=HTTPStatus.UNAUTHORIZED,
                    detail="User does not have required permission",
                    headers={"WWW-Authenticate": "Bearer"},
                )
        return True



from http import HTTPStatus
from httpx._client import AsyncClient
from json import JSONDecodeError
from logging import Logger
from typing import Annotated, Any, TypeAlias, cast

from fastapi import HTTPException, Path, Request
from fastapi.security import OAuth2AuthorizationCodeBearer
import httpx
from keycloak import KeycloakOpenID
from keycloak.exceptions import KeycloakAuthenticationError, KeycloakPostError
from pydantic import (
    BaseModel,
    Field,
    PlainSerializer,
    PlainValidator,
    SecretStr,
    WithJsonSchema,
)
import requests

from phederation.models import APActor
from phederation.utils import ObjectId
from phederation.utils.base import actor_id_from_username, urljoin
from phederation.utils.exceptions import AuthenticationError, UserError
from phederation.utils.logging import configure_logger
from phederation.utils.settings import KeycloakSettings, PhedSettings

from .actors import ActorManager


def serialize_token(token: SecretStr | None):
    if not token:
        return ""
    return token.get_secret_value()


def deserialize_token(token: Any):  # pyright: ignore[reportAny]
    if isinstance(token, SecretStr):
        return token
    if isinstance(token, str):
        return SecretStr(token)
    return None


class TokenSchema(BaseModel):
    token: str


_schema_token = TokenSchema.model_json_schema() | {"token": "SecretStr"}

"""This allows to serialize the UserInfo class, so that it can be used in FastAPI routes.
It is a security concern, but better than making the access_token a string."""
SecretStrAnnotated: TypeAlias = Annotated[
    SecretStr | None, Field(validate_default=True), WithJsonSchema(_schema_token), PlainSerializer(serialize_token), PlainValidator(deserialize_token)
]


class UserInfo(BaseModel):
    preferred_username: str
    actor_id: ObjectId
    email: str | None = None
    full_name: str | None = None
    access_token: SecretStrAnnotated = None
    groups: list[str] | None = None


class Authentication:
    def __init__(self, settings: PhedSettings, actor_manager: ActorManager | None = None, keycloak_settings_client: KeycloakSettings | None = None):
        self.settings: PhedSettings = settings
        self.keycloak_settings: KeycloakSettings = KeycloakSettings()
        self.keycloak_openid: KeycloakOpenID = KeycloakOpenID(
            server_url=self.settings.domain.keycloak_hostname,
            realm_name=self.keycloak_settings.keycloak_realm,
            client_id=self.keycloak_settings.keycloak_client_id,
            client_secret_key=self.keycloak_settings.keycloak_client_secret.get_secret_value(),
            verify=True,
        )
        self.oauth2_scheme_code: OAuth2AuthorizationCodeBearer = OAuth2AuthorizationCodeBearer(
            tokenUrl=f"{self.settings.domain.keycloak_hostname}/realms/{self.keycloak_settings.keycloak_realm}/protocol/openid-connect/token",
            authorizationUrl=f"{self.settings.domain.keycloak_hostname}/realms/{self.keycloak_settings.keycloak_realm}/protocol/openid-connect/auth",
            refreshUrl=f"{self.settings.domain.keycloak_hostname}/realms/{self.keycloak_settings.keycloak_realm}/protocol/openid-connect/token",
            scopes={"openid": "Default openid scope"},
            auto_error=False,  # optional authentication, for routes where the result is "Public" items if somebody is not authenticated
        )
        if keycloak_settings_client:
            self.keycloak_settings = keycloak_settings_client
            self.keycloak_openid = KeycloakOpenID(
                server_url=settings.domain.keycloak_hostname,
                realm_name=keycloak_settings_client.keycloak_realm,
                client_id=keycloak_settings_client.keycloak_client_id,
                client_secret_key=keycloak_settings_client.keycloak_client_secret.get_secret_value(),
                verify=True,
            )
        self.actor_manager: ActorManager | None = actor_manager

        # this can be changed in unit tests to mock url requests
        self.requests: AsyncClient = httpx.AsyncClient()

        self.logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix)

    async def __call__(
        self,
        request: Request,
    ) -> UserInfo | None:
        """
        Authentication endpoint that requires a valid token for elevated access.
        """
        token = await self.oauth2_scheme_code(request)
        self.logger.debug(f"Calling authentication endpoint, token={"***" if token and len(token) > 0 else "(not set)"}")
        if not token:
            return None
        user_info = await self.verify_token(SecretStr(token))
        return user_info

    async def verify_local_username(
        self, username: Annotated[str, Path(description="Username of the actor on this instance, without the URL.")]
    ) -> APActor:
        """
        Checks if the actor with the given username is local to this instance.

        Args:
            username (str): username to verify.

        Returns:
            APActor object.

        Raises:
            UserError if the actor does not exist locally.
        """
        if not self.actor_manager:
            raise AuthenticationError(f"User '{username}' not found: actor_manager is not set (probably because this is not an instance)")

        actor_id = actor_id_from_username(self.settings.domain.base_url, username=username)
        if not actor_id:
            raise AuthenticationError(f"User '{username}' not found: actor_id is None")
        actor = await self.actor_manager.actor_from_storage(id=actor_id)
        if not actor:
            raise UserError(f"User with actor id '{actor_id}' not found on this instance")
        return actor

    def get_openid_config(self) -> dict[str, str]:
        return self.keycloak_openid.well_known()  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

    def authenticate_user(self, username: str, password: SecretStr, redirect_uri: str | None = None) -> SecretStr:
        """
        Authenticate the user using Keycloak and return an access token.
        """
        try:
            if not redirect_uri:
                redirect_uri = self.settings.domain.base_url
            token: dict[str, Any] | None = self.keycloak_openid.token(  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
                username, password.get_secret_value(), redirect_uri=redirect_uri
            )
            if not token:
                raise KeycloakAuthenticationError("Unauthorized, token not available")
            access_token = token.get("access_token", None)
            if not isinstance(access_token, str):
                raise KeycloakAuthenticationError("Unauthorized, access_token not available")
            return SecretStr(access_token)
        except KeycloakAuthenticationError as e:
            self.logger.warning(f"Keycloak: login unsuccessful; KeycloakAuthenticationError error: {e}.")
            raise HTTPException(
                status_code=HTTPStatus.UNAUTHORIZED,
                detail="Unauthorized",
            )
        except KeycloakPostError as e:
            self.logger.warning(f"Keycloak: login unsuccessful; KeycloakPostError error: {e}.")
            raise HTTPException(
                status_code=HTTPStatus.BAD_REQUEST,
                detail="POST/Client error",
            )

    async def verify_token(self, token: SecretStr | None) -> UserInfo | None:
        """
        Verify the given token and return user information if it is valid, None if it is not.
        """
        try:
            if token is None:
                return None

            keycloak_url: str = self.settings.domain.keycloak_docker_hostname
            realm_name: str = self.keycloak_openid.realm_name
            headers = cast(dict[str, Any], self.keycloak_openid.connection.headers)
            headers["Authorization"] = f"Bearer {token.get_secret_value()}"
            complete_url = urljoin(keycloak_url, "realms", realm_name, "protocol/openid-connect/userinfo")

            response: requests.Response | httpx.Response
            try:
                self.logger.info(f"Sending token authorization to url {complete_url}")
                response = await self.requests.get(url=complete_url, headers=headers, follow_redirects=True)
            except (httpx.ConnectError, requests.ConnectionError) as e:
                self.logger.error(f"Could not connect to keycloak, error: {e}. Url: {complete_url}")
                return None
            if response.status_code == HTTPStatus.UNAUTHORIZED:
                return None

            try:
                user_info = cast(dict[str, Any], response.json())
                # user_info = self.keycloak_openid.userinfo(token) # this does not connect to the internal keycloak inside docker
            except JSONDecodeError as e:
                self.logger.error(
                    f"Authentication: JSONDecodeError, status: {response.status_code}",
                    extra={"status_code": response.status_code, "content": response.content.decode("utf-8")},
                )
                return None

            if not user_info:
                self.logger.error(f"Authentication: Invalid token.")
                return None

            if not "preferred_username" in user_info:
                self.logger.error(f"Authentication: Invalid data received from keycloak: preferred_username is not set (token={token}).")
                return None

            self.logger.debug(f"New user info received for user '{user_info.get("preferred_username")}'")
            preferred_username = user_info.get("preferred_username", None)
            if not isinstance(preferred_username, str):
                return None
            user_info = UserInfo(
                preferred_username=preferred_username,
                email=user_info.get("email"),
                full_name=user_info.get("name") or user_info.get("full_name"),
                groups=user_info.get("groups"),
                access_token=token,
                actor_id=actor_id_from_username(base_url=self.settings.domain.base_url, username=preferred_username),
            )

            if self.actor_manager and self.settings.security.create_actor_from_token:
                # check if the user already exists. If not, create them locally.
                username = user_info.preferred_username
                try:
                    _ = await self.verify_local_username(username=username)
                except UserError:
                    self.logger.info(f"Creating local user '{username}' because they successfully logged into keycloak")
                    await self.actor_manager.create_actor_from_username(username)
            
            return user_info
        except KeycloakAuthenticationError as e:
            self.logger.warning(f"Keycloak: login unsuccessful; error: {e}.")
            return None



"""
Announce activity handler implementation.
"""

from typing import override

from phederation.models import dereference
from phederation.models.activities import APActivity
from phederation.models.objects import APObject
from phederation.utils import ObjectId
from phederation.utils.base import AccessType, collection_id_from_name
from phederation.utils.exceptions import HandlerError, ValidationError

from .base import ActivityHandler


class BlockHandler(ActivityHandler):
    """Handle Block actors and objects."""

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """Validate Block activity."""
        # Validate basic structure
        if activity.type != "Block":
            raise ValidationError("Invalid activity type")

        if not activity.object:
            raise ValidationError("Missing object")

        if not activity.actor:
            raise ValidationError("Missing actor")

        # Validate object exists
        object_id = activity.object.id if isinstance(activity.object, APObject) else activity.object
        obj = await self.resolver.resolve_object(object_id)
        if not obj:
            raise ValidationError(f"Object not found: {object_id}")

        # Validate actor can block
        actor = await self.resolver.resolve_actor(activity.actor)
        if not actor:
            raise ValidationError(f"Actor not found: {activity.actor}")

        return activity

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """Process Block activity."""

        # Update object/actor blocks collection
        object_id = dereference(activity, key="object")
        actor_id = dereference(activity, key="actor")
        if not object_id or not actor_id:
            raise HandlerError(f"Block handler failed because object_id={object_id}, actor_id={actor_id}")
        # TODO: make blocks private by default, settings?
        access = AccessType.from_visibility(activity.visibility)
        await self._update_blocks_collection(object_id=object_id, actor_id=actor_id, access=access)

        return activity

    async def _update_blocks_collection(self, object_id: ObjectId, actor_id: ObjectId, access: AccessType) -> None:
        """Update object's blocks collection."""
        collection_blocks = collection_id_from_name(id=actor_id, name='blocks')
        _ = await self.collections.add_to_collection(collection_id=collection_blocks, items=object_id, access=access)

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        # Update object/actor blocks collection, if required. The object will not be a local one, but we may want to update local copies of it
        object_id = dereference(activity, key="object")
        actor_id = dereference(activity, key="actor")
        if not object_id or not actor_id:
            raise HandlerError(f"Block handler inbox failed because object_id={object_id}, actor_id={actor_id}")
        access = AccessType.from_visibility(activity.visibility)
        await self._update_blocks_collection(object_id=object_id, actor_id=actor_id, access=access)



"""ActivityPub models.

Includes the models defined in ActivityPub, such as APObject, APActor, APDocument and all activities, such as APCreate, APBlock, and so on.
"""

from abc import ABCMeta, abstractmethod
from typing import Any, ClassVar, Generic, TypeVar

from pydantic import BaseModel, ConfigDict

from phederation.utils import ActivityType, NodeInfo, ObjectType
from phederation.utils.base import AccessType, ActivityPubBaseWithId, ObjectId
from phederation.utils.exceptions import StorageError, ValidationError, catch_exceptions

from .activities import (
    APAccept,
    APActivity,
    APAnnounce,
    APBlock,
    APCreate,
    APDelete,
    APFollow,
    APLike,
    APMigrate,
    APReject,
    APRemove,
    APAdd,
    APUndo,
    APUpdate,
    ActivityInQueue,
    ErrorMessageActivityHandler,
    ErrorMessageActivityNotFound,
    ErrorMessageActivityRestricted,
    ErrorMessageActivityValidation,
)
from .actors import (
    APAccount,
    APActor,
    APApplication,
    APGroup,
    APOrganization,
    APPerson,
    APService,
    APUser,
    ActorType,
    ErrorMessageActorUnauthorized,
)
from .collections import (
    APCollection,
    APCollectionPage,
    APOrderedCollection,
    APOrderedCollectionPage,
    APUsersCollection,
    ValidCollection,
)
from .keys import APPrivateKey, APPublicKey, ErrorMessageKeyNotFound
from .links import APLink, APMention, RelativeLink, RelativeLinkTemplate
from .maintenance import (
    APMaintenance,
    APMaintenanceCommand,
    APMaintenanceMode,
    APMaintenanceModeCancel,
    MaintenanceCommandType,
)
from .media import MediaMetadata
from .objects import (
    APArticle,
    APAudio,
    APCollectionItem,
    APDocument,
    APEvent,
    APImage,
    APNote,
    APObject,
    APPage,
    APPlace,
    APProfile,
    APRelationship,
    APTombstone,
    APUserData,
    APVideo,
    DocumentType,
    dereference,
    dereference_or_raise,
)
from .proofs import DataIntegrityProof

T = TypeVar("T", bound=ActivityPubBaseWithId)


class CRUD(BaseModel, Generic[T], metaclass=ABCMeta):
    """Base class for CRUD operations."""

    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    async def get_table_name(self) -> str:
        pass

    @catch_exceptions(StorageError, "Failed to create_many")
    @abstractmethod
    async def create_many(self, data: list[T]) -> list[ObjectId]:
        pass

    @catch_exceptions(StorageError, "Failed to create")
    @abstractmethod
    async def create(self, data: T, raise_if_exists: bool = True) -> ObjectId:
        pass

    @catch_exceptions(StorageError, "Failed to read")
    @abstractmethod
    async def read(self, id: ObjectId) -> T | None:
        pass

    @catch_exceptions(StorageError, "Failed to select")
    @abstractmethod
    async def select(
        self, collection_id: ObjectId | None = None, stored_ids: list[ObjectId] | None = None, access: AccessType = AccessType.PUBLIC
    ) -> list[T]:
        pass

    @catch_exceptions(StorageError, "Failed to count")
    @abstractmethod
    async def count(self, collection_id: ObjectId | None = None, access: AccessType = AccessType.PUBLIC) -> int:
        pass

    @catch_exceptions(StorageError, "Failed to upsert")
    @abstractmethod
    async def upsert(self, id: ObjectId, data: T) -> ObjectId:
        pass

    @catch_exceptions(StorageError, "Failed to delete")
    @abstractmethod
    async def delete(self, id: ObjectId | list[ObjectId]) -> bool:
        pass


type_to_obj_dict = {
    ActivityType.ACCEPT.value: APAccept,
    ActivityType.ANNOUNCE.value: APAnnounce,
    ActivityType.BLOCK.value: APBlock,
    ActivityType.CREATE.value: APCreate,
    ActivityType.DELETE.value: APDelete,
    ActivityType.FOLLOW.value: APFollow,
    ActivityType.LIKE.value: APLike,
    ActivityType.REJECT.value: APReject,
    ActivityType.REMOVE.value: APRemove,
    ActivityType.ADD.value: APAdd,
    ActivityType.UNDO.value: APUndo,
    ActivityType.UPDATE.value: APUpdate,
    ActivityType.MIGRATE.value: APMigrate,
    ActivityType.ACTIVITY_IN_QUEUE.value: ActivityInQueue,
    ActorType.APPLICATION.value: APApplication,
    ActorType.GROUP.value: APGroup,
    ActorType.ORGANIZATION.value: APOrganization,
    ActorType.PERSON.value: APPerson,
    ActorType.SERVICE.value: APService,
    ActorType.ACTOR.value: APActor,
    ActorType.ACCOUNT.value: APAccount,
    ActorType.PROFILE.value: APProfile,
    ActorType.USER.value: APUser,
    ActorType.PERSON.value: APPerson,
    ActorType.APPLICATION.value: APApplication,
    ActorType.ORGANIZATION.value: APOrganization,
    ActorType.GROUP.value: APGroup,
    ActorType.SERVICE.value: APService,
    DocumentType.ARTICLE.value: APArticle,
    DocumentType.AUDIO.value: APAudio,
    DocumentType.VIDEO.value: APVideo,
    DocumentType.DOCUMENT.value: APDocument,
    DocumentType.IMAGE.value: APImage,
    DocumentType.NOTE.value: APNote,
    DocumentType.USERDATA.value: APUserData,
    ObjectType.COLLECTION.value: APCollection,
    ObjectType.COLLECTION_PAGE.value: APCollectionPage,
    ObjectType.ORDERED_COLLECTION.value: APOrderedCollection,
    ObjectType.ORDERED_COLLECTION_PAGE.value: APOrderedCollectionPage,
    ObjectType.COLLECTION_ITEM.value: APCollectionItem,
    ObjectType.OBJECT.value: APObject,
    ObjectType.PAGE.value: APPage,
    ObjectType.LINK.value: APLink,
    ObjectType.MENTION.value: APMention,
    ObjectType.EVENT.value: APEvent,
    ObjectType.PLACE.value: APPlace,
    ObjectType.RELATIONSHIP.value: APRelationship,
    ObjectType.KEY.value: APPrivateKey,
    ObjectType.DATA_INTEGRITY_PROOF.value: DataIntegrityProof,
    ObjectType.TOMBSTONE.value: APTombstone,
    ObjectType.NODE_INFO.value: NodeInfo,
    ActivityType.MAINTENANCE.value: APMaintenance,
    MaintenanceCommandType.MAINTENANCE_MODE.value: APMaintenanceMode,
    MaintenanceCommandType.MAINTENANCE_MODE_CANCEL.value: APMaintenanceModeCancel,
}


def object_from_type(obj: APObject | dict[str, Any]):
    object_type = dereference_or_raise(obj, "type")
    if isinstance(obj, APObject):
        obj = obj.serialize()
    model_class = type_to_obj_dict.get(object_type, None)
    if model_class:
        obj_new = model_class.deserialize(obj)
    else:
        raise ValidationError(f"Not able to deserialize APObject of type {object_type}")
    return obj_new


# Export all classes
__all__ = [
    "CRUD",
    "T",
    "MediaMetadata",
    "dereference",
    "APPublicKey",
    "APPrivateKey",
    "ActivityPubBaseWithId",
    "APObject",
    "APLink",
    "APActivity",
    "RelativeLink",
    "RelativeLinkTemplate",
    "APActor",
    "ActorType",
    "APUser",
    "APPerson",
    "APGroup",
    "APOrganization",
    "APApplication",
    "APService",
    "APAccount",
    "APEvent",
    "APPlace",
    "APProfile",
    "APRelationship",
    "APTombstone",
    "APArticle",
    "APAudio",
    "APDocument",
    "APImage",
    "APNote",
    "APPage",
    "APVideo",
    "APCollection",
    "APOrderedCollection",
    "APCollectionPage",
    "APOrderedCollectionPage",
    "ValidCollection",
    "APUsersCollection",
    "APCreate",
    "APUpdate",
    "APDelete",
    "APFollow",
    "APUndo",
    "APLike",
    "APAnnounce",
    "APMention",
    "APAccept",
    "APAdd",
    "APRemove",
    "APBlock",
    "APReject",
    "APMigrate",
    "APMaintenanceCommand",
    "APMaintenanceModeCancel",
    "APUserData",
    "APMaintenance",
    "MaintenanceCommandType",
    "ErrorMessageActivityNotFound",
    "ErrorMessageActivityRestricted",
    "ErrorMessageKeyNotFound",
    "ErrorMessageActivityValidation",
    "ErrorMessageActivityHandler",
    "ErrorMessageActorUnauthorized",
    "DataIntegrityProof",
]



"""
collections.py
This module defines the Collection types for ActivityPub.
These collections are based on the ActivityStreams vocabulary and ActivityPub protocol.

For more information on Collections in ActivityStreams, see:
https://www.w3.org/TR/activitystreams-core/#collections
"""

from collections.abc import Sequence
from enum import Enum
from typing import ClassVar

from pydantic import Field
from pydantic.config import ConfigDict

from phederation.utils.base import ObjectId

from .objects import APObject


class ValidCollection(str, Enum):
    Outbox = "outbox"
    Inbox = "inbox"
    Following = "following"
    Followers = "followers"
    Shares = "shares"
    Likes = "likes"
    Liked = "liked"
    Objects = "objects"
    NodeInfo = "nodeinfo"
    Blocks = "blocks"
    Keys = "keys"
    Users = "users"
    Collections = "collections"


class APCollection(APObject):
    """
    Represents a Collection in ActivityPub.

    A Collection is an unordered set of Objects or Links.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#collection
    Usage: https://www.w3.org/TR/activitypub/#collections
    """

    type: str = Field(default="Collection", description="Indicates a Collection")
    total_items: None | int = Field(default=None, ge=0, description="Total number of items")  # Greater than or equal to 0
    current: "str | ObjectId | APCollectionPage | APOrderedCollectionPage | None" = Field(default=None, description="Current page")
    first: "str | ObjectId | APCollectionPage | APOrderedCollectionPage | None" = Field(default=None, description="First page")
    last: "str | ObjectId | APCollectionPage | APOrderedCollectionPage | None" = Field(default=None, description="Last page")
    items: None | Sequence[str | ObjectId | APObject] = Field(default=None, description="Items in the collection")

    model_config: ClassVar[ConfigDict] = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "id": "string",
                    "type": "Collection",
                    "attachment": [],
                    "attributedTo": "string",
                    "content": "string",
                    "context": "string",
                    "name": "string",
                    "endTime": "2025-11-09T21:36:53.227Z",
                    "generator": "string",
                    "icon": "string",
                    "image": "string",
                    "inReplyTo": "string",
                    "location": "string",
                    "preview": "string",
                    "published": "2025-11-09T21:36:53.227Z",
                    "replies": "string",
                    "startTime": "2025-11-09T21:36:53.227Z",
                    "summary": "string",
                    "tag": [],
                    "updated": "2025-11-09T21:36:53.228Z",
                    "url": ["string"],
                    "audience": ["string"],
                    "to": ["string"],
                    "bto": ["string"],
                    "cc": ["string"],
                    "bcc": ["string"],
                    "mediaType": "string",
                    "visibility": "public",
                    "likes": "string",
                    "shares": "string",
                    "totalItems": 0,
                    "current": "string",
                    "first": "string",
                    "last": "string",
                    "items": [],
                }
            ]
        }
    )


class APOrderedCollection(APCollection):
    """
    Represents an OrderedCollection in ActivityPub.

    An OrderedCollection is a Collection where the items are strictly ordered.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#orderedcollection
    Usage: https://www.w3.org/TR/activitypub/#ordered-collections
    """

    type: str = Field(default="OrderedCollection", description="Indicates an OrderedCollection")
    ordered_items: None | Sequence[str | ObjectId | APObject] = Field(default=None, description="The ordered items")

    model_config: ClassVar[ConfigDict] = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "id": "string",
                    "type": "OrderedCollection",
                    "attachment": [],
                    "attributedTo": "string",
                    "content": "string",
                    "context": "string",
                    "name": "string",
                    "endTime": "2025-11-09T21:36:53.227Z",
                    "generator": "string",
                    "icon": "string",
                    "image": "string",
                    "inReplyTo": "string",
                    "location": "string",
                    "preview": "string",
                    "published": "2025-11-09T21:36:53.227Z",
                    "replies": "string",
                    "startTime": "2025-11-09T21:36:53.227Z",
                    "summary": "string",
                    "tag": [],
                    "updated": "2025-11-09T21:36:53.228Z",
                    "url": ["string"],
                    "audience": ["string"],
                    "to": ["string"],
                    "bto": ["string"],
                    "cc": ["string"],
                    "bcc": ["string"],
                    "mediaType": "string",
                    "visibility": "public",
                    "likes": "string",
                    "shares": "string",
                    "totalItems": 0,
                    "current": "string",
                    "first": "string",
                    "last": "string",
                    "orderedItems": [],
                }
            ]
        }
    )


class APCollectionPage(APCollection):
    """
    Represents a CollectionPage in ActivityPub.

    A CollectionPage represents a single page of a Collection.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#collectionpage
    Usage: https://www.w3.org/TR/activitypub/#collection-paging
    """

    type: str = Field(default="CollectionPage", description="Indicates a CollectionPage")
    part_of: "None | str | ObjectId | APCollection" = Field(default=None, description="The parent collection")
    next: "None | str | ObjectId | APCollectionPage" = Field(default=None, description="The next page")
    prev: "None | str | ObjectId | APCollectionPage" = Field(default=None, description="The previous page")


class APOrderedCollectionPage(APCollectionPage, APOrderedCollection):
    """
    Represents an OrderedCollectionPage in ActivityPub.

    An OrderedCollectionPage represents a single page of an OrderedCollection.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#orderedcollectionpage
    Usage: https://www.w3.org/TR/activitypub/#ordered-collection-paging
    """

    type: str = Field(default="OrderedCollectionPage", description="Indicates an OrderedCollectionPage")
    start_index: None | int = Field(default=None, ge=0, description="The index of the first item")  # Greater than or equal to 0


class APFollowersCollection(APOrderedCollection):
    """Collection of followers."""

    type: str = Field(default="OrderedCollection", description="Indicates an OrderedCollection")
    name: str | None = Field(default="Followers")


class APFollowingCollection(APOrderedCollection):
    """Collection of following."""

    type: str = Field(default="OrderedCollection", description="Indicates an OrderedCollection")
    name: str | None = Field(default="Following")


class APLikedCollection(APOrderedCollection):
    """Collection of liked objects."""

    type: str = Field(default="OrderedCollection", description="Indicates an OrderedCollection")
    name: str | None = Field(default="Liked")


class APSharedCollection(APOrderedCollection):
    """Collection of shared objects."""

    type: str = Field(default="OrderedCollection", description="Indicates an OrderedCollection")
    name: str | None = Field(default="Shared")


class APUsersCollection(APOrderedCollection):
    """Collection of all users."""

    type: str = Field(default="OrderedCollection", description="Indicates an OrderedCollection")
    name: str | None = Field(default="Users")

    model_config: ClassVar[ConfigDict] = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "id": "https://example.org/users",
                    "type": "OrderedCollection",
                    "items": [
                        {
                            "id": "https://example.org/users/exampleuser",
                            "type": "Person",
                            "username": "exampleuser",
                            "domain": "example.org",
                            "inbox": "https://example.org/users/exampleuser/inbox",
                            "outbox": "https://example.org/users/exampleuser/outbox",
                            "following": "https://example.org/users/exampleuser/following",
                            "followers": "https://example.org/users/exampleuser/followers",
                            "autoaccept": "True",
                            "shared_inbox": "https://example.org/inbox",
                            "keyId": "https://example.org/keys/c7d3075a-69c8-47fb-92b0-288898d2f015",
                            "createdAt": "2023-01-19 12:44:48.817540",
                            "updatedAt": "None",
                            "lastFetchedAt": "None",
                        }
                    ],
                    "@context": [
                        "https://www.w3.org/ns/activitystreams",
                        "https://w3id.org/security/v1",
                    ],
                }
            ]
        }
    )



"""
Base activity handler implementation.
"""

from abc import ABC, abstractmethod
from logging import Logger

from phederation.collections.collections import CollectionManager
from phederation.federation.delivery import ActivityDelivery
from phederation.federation.resolver import ActivityPubResolver
from phederation.models import APActivity
from phederation.security.key_management import KeyManager
from phederation.security.actors import ActorManager
from phederation.storage.base import StorageBackend
from phederation.utils.logging import configure_logger


class ActivityHandler(ABC):
    """Base activity handler."""

    def __init__(
        self,
        actor_manager: ActorManager,
        key_manager: KeyManager,
        storage: StorageBackend,
        collections: CollectionManager,
        resolver: ActivityPubResolver,
        delivery: ActivityDelivery,
    ):
        self.actor_manager: ActorManager = actor_manager
        self.key_manager: KeyManager = key_manager
        self.storage: StorageBackend = storage
        self.collections: CollectionManager = collections
        self.resolver: ActivityPubResolver = resolver
        self.delivery: ActivityDelivery = delivery
        self.logger: Logger = configure_logger(
            f"{".".join(__name__.split('.')[:-1])}.{type(self).__name__.lower()}", prefix=self.storage.settings.federation.logging_prefix
        )

    @abstractmethod
    async def validate(self, activity: APActivity) -> APActivity:
        """
        Validate activity.

        Args:
            activity: Activity to validate

        Raises:
            ValidationError if validation fails

        Returns:
            The activity dictionary, potentially changed due to validation.
        """
        pass

    @abstractmethod
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """
        Process activity that was posted into an outbox (so, client to server).

        Args:
            activity: Activity to process

        Returns:
            Activity ID if successful

        Raises:
            HandlerError if processing fails
        """
        pass

    @abstractmethod
    async def process_inbox(self, activity: APActivity) -> None | str:
        """
        Process activity that was posted into an inbox (so, server to server).

        Args:
            activity: Activity to process

        Returns:
            Activity ID if successful

        Raises:
            HandlerError if processing fails
        """
        pass



"""
NoSQL storage backend CRUD operations for ActivityPub data.
"""

from collections.abc import Awaitable, Callable
from logging import Logger
from typing import Any, cast, override

from pydantic import Field, field_serializer
from pymongo.asynchronous.collection import AsyncCollection
from pymongo.asynchronous.cursor import AsyncCursor

from phederation.models import CRUD, T, object_from_type
from phederation.utils import ObjectId
from phederation.utils.base import AccessType, UrlType, assemble_id_url
from phederation.utils.exceptions import StorageError, catch_exceptions


def nosql_build_search_query(
    collection_id: ObjectId | None, ids: list[ObjectId] | None = None, access: AccessType = AccessType.PUBLIC
) -> dict[str, list[dict[str, ObjectId] | dict[str, list[dict[str, str]]]]]:
    query: dict[str, list[dict[str, ObjectId | str] | dict[str, list[dict[str, str]]]]] | None = {"$and": []}
    if collection_id is not None:
        query["$and"].append({"collectionId": collection_id})
    if ids is not None:
        query["$and"].append({"$or": [{"storedId": id} for id in ids]})

    if access == AccessType.PUBLIC:
        query["$and"].append({"visibility": AccessType.PUBLIC.value})
    elif access == AccessType.FOLLOWER:
        query["$and"].append({"$or": [{"visibility": AccessType.PUBLIC.value}, {"visibility": AccessType.FOLLOWER.value}]})
    elif access == AccessType.PRIVATE:
        pass  # use all items
    if len(query["$and"]) == 0:
        query = {}
    return query


class NoSqlCRUD(CRUD[T]):
    get_collection: Callable[[str], Awaitable[AsyncCollection[Any]]]
    table_name: str
    base_url: ObjectId
    logger: Logger = Field(default=..., exclude=True)
    MODEL: type

    @field_serializer("logger")
    def serialize_logger(self, logger: Logger):
        logger = logger
        return "None"

    @override
    async def get_table_name(self) -> str:
        return self.table_name

    @catch_exceptions(StorageError, "NoSQL: Failed to create_many")
    @override
    async def create_many(self, data: list[T]) -> list[ObjectId]:
        data_dict = [d.serialize(include_context=False) for d in data]
        object_ids: list[ObjectId | str | None] = []
        for d in data_dict:
            if "id" in d:
                object_id = d.get("id")
                d["_id"] = object_id  # necessary for mongodb
            else:
                object_id = None

            object_ids.append(object_id)
        collection = await self.get_collection(self.table_name)
        new_items = await collection.insert_many(data_dict)

        # update the id string according to the new primary key
        new_object_ids: list[ObjectId | str] = []
        for inserted_id, object_id in zip(cast(list[ObjectId], new_items.inserted_ids), object_ids):
            if object_id is None:
                object_id = assemble_id_url(type=UrlType(self.table_name), primary=(inserted_id), base_url=self.base_url)
                query_filter = {"_id": (inserted_id)}
                update_operation = {"$set": {"id": object_id}}
                _ = await collection.update_one(query_filter, update_operation)
            new_object_ids.append(object_id)
        return [object_id for object_id in object_ids if object_id]

    @catch_exceptions(StorageError, "NoSQL: Failed to create")
    @override
    async def create(self, data: T, raise_if_exists: bool = True) -> ObjectId:
        data_dict = data.serialize(include_context=False)
        if "id" in data_dict:
            object_id = data_dict.get("id")
            data_dict["_id"] = object_id  # necessary for mongodb
        else:
            object_id = None

        collection = await self.get_collection(self.table_name)

        # if there is already an item in the collection with the given id, change the id and insert
        try:
            new_item = await collection.insert_one(data_dict)
        except Exception as e:
            # this can happen if an instance receives a message for multiple actors.
            # TODO: this is not a very efficient way to resolve multiple messages
            self.logger.warning(f"Data '{data}' with id '{object_id}' already exists in storage")
            if raise_if_exists or not object_id:
                raise e
            else:
                return cast(str, object_id)

        # update the id string according to the new primary key
        if object_id is None:
            new_inserted_id = cast(ObjectId, new_item.inserted_id)
            object_id = assemble_id_url(type=UrlType(self.table_name), primary=new_inserted_id, base_url=self.base_url)
            query_filter = {"_id": new_inserted_id}
            update_operation = {"$set": {"id": object_id}}
            _ = await collection.update_one(query_filter, update_operation)
        return object_id

    @catch_exceptions(StorageError, "NoSQL: Failed to read")
    @override
    async def read(self, id: ObjectId) -> T | None:
        collection = await self.get_collection(self.table_name)
        result = cast(dict[str, Any], await collection.find_one({"id": id}))

        if result:
            result.pop("_id", None)
            obj = cast(T, object_from_type(self.MODEL.deserialize(result)))  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
            return obj
        else:
            return None

    @catch_exceptions(StorageError, "NoSQL: Failed to select")
    @override
    async def select(
        self, collection_id: ObjectId | None = None, stored_ids: list[ObjectId] | None = None, access: AccessType = AccessType.PUBLIC
    ) -> list[T]:
        collection = await self.get_collection(self.table_name)
        results: list[T] = []
        query = nosql_build_search_query(collection_id=collection_id, ids=stored_ids, access=access)
        async with collection.find(query) as cursor:
            async for result in cast(AsyncCursor[dict[str, Any]], cursor):
                if result:
                    result.pop("_id", None)
                    obj = cast(
                        T, object_from_type(self.MODEL.deserialize(result))  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                    )
                    results.append(obj)
        return results

    @catch_exceptions(StorageError, "Failed to count")
    @override
    async def count(self, collection_id: ObjectId | None = None, access: AccessType = AccessType.PUBLIC) -> int:
        collection = await self.get_collection(self.table_name)
        query = nosql_build_search_query(collection_id=collection_id, access=access)
        total_items = await collection.count_documents(query)
        return total_items

    @catch_exceptions(StorageError, "NoSQL: Failed to update")
    @override
    async def upsert(self, id: ObjectId, data: T) -> ObjectId:
        """Upsert an object with given id.
        If the id cannot be found, inserts the new object with the .id field set to the parameter id.

        Args:
            id (ObjectId): Id of the object to update.
            data (T): Object information.

        Returns:
            ObjectId: The input id.
        """
        collection = await self.get_collection(self.table_name)
        query_filter = {"id": id}
        data.id = id
        data_dict = data.serialize(include_context=False)
        update_operation = {"$set": data_dict}
        _ = await collection.update_one(query_filter, update_operation, upsert=True)
        return id

    @catch_exceptions(StorageError, "NoSQL: Failed to delete")
    @override
    async def delete(self, id: ObjectId | list[ObjectId]) -> bool:
        collection = await self.get_collection(self.table_name)
        if isinstance(id, str):
            _ = await collection.delete_one({"id": id})
        else:
            if len(id) > 0:
                filter = {"$or": [{"id": i} for i in id]}
                _ = await collection.delete_many(filter)
        return True



import asyncio
import logging.config
from typing import Any, cast
import uuid

from temporalio import workflow
from temporalio.worker import Worker
from temporalio.worker.workflow_sandbox import (
    SandboxRestrictions,
    SandboxedWorkflowRunner,
)
import yaml

with workflow.unsafe.imports_passed_through():

    from phederation.storage.tasks import StorageBackgroundTask
    from phederation.actions.workflows import (
        ActivityDeliveryWorkflow,
        DeliveryWorkerTask,
        UpdateNodeInfoTask,
        KeyRotationTask,
    )
    from phederation.federation.server import ActivityPubServer
    from phederation.utils.settings import PhedSettings
    from phederation.actions.workflows import UpdateNodeInfoWorkflow, KeyRotationWorkflow

interrupt_event = asyncio.Event()


async def main_worker(settings_file: str | None = None, logging_file: str | None = None) -> None:
    settings_file = settings_file or "phederation.yaml"
    settings: PhedSettings = PhedSettings(yaml_file=settings_file)
    if not settings.worker.enabled or not settings.worker.start_own_worker:
        print("Worker disabled in settings, shutting down.")
        return

    server: ActivityPubServer = ActivityPubServer(settings=settings)

    if logging_file is None:
        with open(settings_file) as f:
            settings_config = cast(dict[str, Any], yaml.safe_load(f))
            logging_config = cast(dict[str, Any], settings_config["logging"])
    else:
        with open(logging_file) as f:
            logging_config = cast(dict[str, Any], yaml.safe_load(f))
    logging.config.dictConfig(config=logging_config)

    await server.initialize()

    if not server.worker_client:
        print("Worker could not be started: Client not connected.")
        return

    storage_task = StorageBackgroundTask()
    upgrade_node_info_task = UpdateNodeInfoTask(settings.federation.update_nodeinfo_period)
    delivery_worker = DeliveryWorkerTask(settings.federation.delivery_queue_polling_period)
    key_rotation_task = KeyRotationTask(enabled=settings.security.key_rotation_enabled, update_period=1000)

    await storage_task.initialize(server=server)
    await upgrade_node_info_task.initialize(server=server)
    await delivery_worker.initialize(server=server)
    await key_rotation_task.initialize(server=server)

    # Run a worker for the workflow
    async with Worker(
        server.worker_client,
        task_queue=server.settings.worker.tasks_queue_name,
        identity=f"gunicorn-worker-{str(uuid.uuid4())[:8]}",
        workflows=[UpdateNodeInfoWorkflow, ActivityDeliveryWorkflow, KeyRotationWorkflow],
        workflow_runner=SandboxedWorkflowRunner(
            restrictions=SandboxRestrictions.default.with_passthrough_modules("pydantic", "annotated_types", "pydantic_core")
        ),
        activities=[
            upgrade_node_info_task.update_node_info,
            delivery_worker.poll_delivery_queue,
            delivery_worker.prepare_inboxes_for_delivery,
            delivery_worker.deliver_activity,
            key_rotation_task.rotate_keys,
            key_rotation_task.announce_key_rotation,
        ],
    ):
        # Wait until interrupted
        print("Worker started, ctrl+c to exit")
        _ = await interrupt_event.wait()
        print("Shutting down")


if __name__ == "__main__":
    loop = asyncio.new_event_loop()
    try:
        loop.run_until_complete(main_worker())
    except KeyboardInterrupt:
        interrupt_event.set()
        loop.run_until_complete(loop.shutdown_asyncgens())



from dataclasses import dataclass
from datetime import datetime, timezone
from http import HTTPStatus
from logging import Logger
from typing import Any, cast
from urllib.parse import urlparse

import httpx

from phederation.cache import BaseCache, with_cache
from phederation.cache.base import WithCache
from phederation.utils import NodeInfo, ObjectId
from phederation.utils.base import APDateTime, InstanceEndpoints, urljoin
from phederation.utils.exceptions import DiscoveryError, catch_exceptions
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings
from phederation.utils.version import PHEDERATION_USERAGENT


@dataclass
class InstanceInfo:
    """Information about the instance software. Shared across the network."""

    domain: str
    nodeinfo: None | NodeInfo
    software_version: None | str
    instance_actor: None | dict[str, Any]
    shared_inbox: None | ObjectId
    endpoints: dict[str, str] | InstanceEndpoints
    features: dict[str, bool]
    last_updated: APDateTime

    @staticmethod
    def from_dict(base_url: str, info_dict: dict[str, Any]):
        nodeinfo_dict = cast(dict[str, Any] | None, info_dict.get("nodeinfo"))
        if not nodeinfo_dict:
            return None
        nodeinfo = NodeInfo.deserialize(nodeinfo_dict)
        if not isinstance(nodeinfo, NodeInfo):
            return None
        instance_actor = cast(dict[str, Any] | None, info_dict.get("instance_actor"))
        shared_inbox = cast(ObjectId | None, info_dict.get("shared_inbox"))
        endpoints = cast(dict[str, str], info_dict.get("endpoints"))
        features = cast(dict[str, bool], info_dict.get("features"))

        return InstanceInfo(
            domain=urlparse(base_url).netloc,
            nodeinfo=nodeinfo,
            software_version=nodeinfo.software.get("version") if nodeinfo else None,
            instance_actor=instance_actor,
            shared_inbox=shared_inbox,
            endpoints=endpoints,
            features=features,
            last_updated=datetime.now(timezone.utc),
        )


class InstanceDiscovery(WithCache):
    """Federation instance discovery.

    Features:
     - Instance metadata discovery
     - WebFinger resolution
     - NodeInfo discovery
     - Actor discovery
    """

    def __init__(self, settings: PhedSettings, cache: BaseCache | None = None, request_timeout: int = 10):  # 1 hour
        """Initialize instance discovery.

        Args:
            settings (PhedSettings): To obtain the base_url of the instance.
            cache_ttl (int, optional): Time-to-live for the internal cache. Defaults to 3600.
            request_timeout (int, optional): How many seconds the discovery waits on remote instances until a timeout. Defaults to 10.
        """
        self.settings: PhedSettings = settings
        self.logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix)
        self.cache: BaseCache | None = cache
        self.timeout: int = request_timeout
        # this can be set by the test suite to replace sending actual http requests and instead use the test_client
        self.requests: httpx.AsyncClient = httpx.AsyncClient()
        self.headers: dict[str, str] = {"User-Agent": PHEDERATION_USERAGENT, "Accept": "application/activity+json"}

    def is_local(self, url: str):
        return str(url).startswith(self.settings.domain.base_url)

    async def initialize(self) -> None:
        """Initialize discovery."""
        pass

    @catch_exceptions(DiscoveryError, "Failed instance discovery")
    @with_cache(__name__ + ".discover_instance")
    async def discover_instance(self, instance_base_url: str) -> InstanceInfo:
        """
        Discover instance information.

        Args:
            instance_base_url: Instance base url (scheme://netloc)

        Returns:
            InstanceInfo with complete instance data
        """

        # Discover instance components.
        # We first try to directly get the instance info directly, then individually try endpoints
        info = await self.discover_info(base_url=instance_base_url)
        if not info:
            nodeinfo = await self.discover_nodeinfo(base_url=instance_base_url)
            instance_actor = await self.discover_instance_actor(base_url=instance_base_url)
            endpoints = await self.discover_endpoints(base_url=instance_base_url)
            features = await self.discover_features(base_url=instance_base_url)

            shared_inbox = endpoints.get("shared_inbox", None)

            # Build instance info
            info = InstanceInfo(
                domain=urlparse(instance_base_url).netloc,
                nodeinfo=nodeinfo,
                software_version=nodeinfo.software.get("version") if nodeinfo else None,
                instance_actor=instance_actor,
                shared_inbox=shared_inbox if shared_inbox else None,
                endpoints=endpoints,
                features=features,
                last_updated=datetime.now(timezone.utc),
            )

        return info

    @catch_exceptions(DiscoveryError, "Failed instance info discovery")
    @with_cache(__name__ + ".discover_info")
    async def discover_info(self, base_url: str):
        well_known_url: ObjectId = urljoin(base_url, "/.well-known/instanceinfo")
        response = await self.requests.get(well_known_url, headers=self.headers)
        if response.status_code != 200:
            raise DiscoveryError(f"Response in discover_info not ok, status_code={response.status_code}, detail={response.content}")

        info_dict = cast(dict[str, Any], response.json())
        instance_info = InstanceInfo.from_dict(base_url=base_url, info_dict=info_dict)
        return instance_info

    @catch_exceptions(DiscoveryError, "Failed node info discovery")
    @with_cache(__name__ + ".discover_nodeinfo")
    async def discover_nodeinfo(self, base_url: str) -> None | NodeInfo:
        """
        Discover NodeInfo data.

        Implements NodeInfo 2.0 and 2.1 discovery.
        """
        # Try well-known location first
        well_known_url: ObjectId = urljoin(base_url, "/.well-known/nodeinfo")
        response = await self.requests.get(well_known_url, headers=self.headers)
        if response.status_code != 200:
            raise DiscoveryError(f"Response not ok, status_code={response.status_code}, detail={response.content}")

        links = cast(dict[str, Any], response.json())
        nodeinfo_url: str | None = None

        # Find highest supported version
        unresolved_links = cast(list[dict[str, str]], links.get("links", []))
        for link in unresolved_links:
            if link.get("rel") == "http://nodeinfo.diaspora.software/ns/schema/2.1":
                nodeinfo_url = link.get("href")
                break
            elif link.get("rel") == "http://nodeinfo.diaspora.software/ns/schema/2.0":
                nodeinfo_url = link.get("href")

        if not nodeinfo_url:
            raise DiscoveryError(f"Response not ok, nodeinfo_url=None, links={links}, detail={response.content}")

        # Fetch NodeInfo
        nodeinfo_response = await self.requests.get(nodeinfo_url, headers=self.headers)
        if nodeinfo_response.status_code != 200:
            raise DiscoveryError(
                f"Response not ok, nodeinfo_url get failed with status_code={nodeinfo_response.status_code}, detail={nodeinfo_response.content}"
            )

        data = cast(dict[str, Any], nodeinfo_response.json())
        nodeinfo = NodeInfo.deserialize(data)
        if not isinstance(nodeinfo, NodeInfo):
            raise DiscoveryError(f"NodeInfo could not be deserialized properly from data: {data}")
        return nodeinfo

    @catch_exceptions(DiscoveryError, "Failed instance actor discovery")
    async def discover_instance_actor(self, base_url: str) -> None | dict[str, Any]:
        """
        Discover an instance actor.

        Tries multiple common endpoints:
            - /actor
            - /instance
            - /instance/actor
            - /

        Args:
            base_url: the url of the instance, without the actor path added.
        """
        locations: list[ObjectId] = [urljoin(base_url, url) for url in ["/actor", "/instance", "/instance/actor", "/"]]

        headers = {"Accept": 'application/ld+json; profile="https://www.w3.org/ns/activitystreams"'}

        for url in locations:
            try:
                response = await self.requests.get(url, headers=headers)
                if response.status_code == HTTPStatus.OK:
                    data = cast(dict[str, Any], response.json())
                    actor_type = data.get("type", None)
                    if isinstance(actor_type, str) and actor_type.lower() in ["application", "service"]:
                        return data
                else:
                    self.logger.warning(
                        f"Could not get a proper response from host instance during actor discovery: status_code={response.status_code}, detail={response.content}"
                    )
            except:
                continue

        return None

    async def discover_endpoints(self, base_url: str) -> dict[str, str]:
        """
        Discover instance endpoints.

        Finds common ActivityPub endpoints.

        Args:
            base_url: the url of the instance, without the actor path added.
        """
        endpoints: dict[str, str] = {}

        # Common endpoint paths
        paths = {
            "inbox": "/inbox",
            "outbox": "/outbox",
            "following": "/following",
            "followers": "/followers",
            "featured": "/featured",
            "shared_inbox": "/inbox",
            "nodeinfo": "/.well-known/nodeinfo",
            "webfinger": "/.well-known/webfinger",
        }

        for name, path in paths.items():
            url: ObjectId = urljoin(base_url, path)
            try:
                response = await self.requests.head(url, headers=self.headers)
                if response.status_code != HTTPStatus.NOT_FOUND:
                    endpoints[name] = url
            except Exception as e:
                self.logger.debug(f"Exception in discover_endpoints: {type(e).__name__}, {e}")
                continue

        return endpoints

    async def discover_features(self, base_url: str) -> dict[str, bool]:
        """
        Discover supported features.

        Checks for various federation features:
         - activitypub protocol
         - webfinger
         - nodeinfo
         - shared inbox
         - collections route

        Args:
            base_url: the url of the instance, without the actor path added.
        """
        features = {"activitypub": False, "webfinger": False, "nodeinfo": False, "shared_inbox": False, "collections": False, "media_proxy": False}

        # Check WebFinger
        try:
            webfinger_url = urljoin(base_url, f"/.well-known/webfinger?resource=acct:test@{base_url}")
            response = await self.requests.head(webfinger_url, headers=self.headers)
            features["webfinger"] = response.status_code != 404
        except:
            pass

        # Check NodeInfo
        try:
            nodeinfo_url = urljoin(base_url, "/.well-known/nodeinfo")
            response = await self.requests.head(nodeinfo_url, headers=self.headers)
            features["nodeinfo"] = response.status_code == 200
        except:
            pass

        # Check shared inbox
        try:
            inbox_url = urljoin(base_url, "/inbox")
            response = await self.requests.head(inbox_url, headers=self.headers)
            features["shared_inbox"] = response.status_code != 404
        except:
            pass

        # Check collections
        try:
            collections = ["/following", "/followers", "/featured"]
            check_paths: list[bool] = [await self._check_endpoint(base_url=base_url, path=path) for path in collections]
            features["collections"] = any(check_paths)
        except:
            pass

        # Check media proxy
        try:
            proxy_url = urljoin(base_url, "/proxy")
            response = await self.requests.head(proxy_url, headers=self.headers)
            features["media_proxy"] = response.status_code != 404
        except:
            pass

        # ActivityPub is supported if basic endpoints exist
        features["activitypub"] = features["shared_inbox"] or features["collections"]

        return features

    async def _check_endpoint(self, base_url: str, path: str) -> bool:
        """Check if endpoint exists."""
        try:
            url = urljoin(base_url, path)
            response = await self.requests.head(url, headers=self.headers)
            return response.status_code != 404
        except:
            return False

    async def webfinger(self, resource: str, domain: None | str = None) -> None | dict[str, Any]:
        """
        Perform WebFinger lookup.

        Args:
            resource: Resource to look up (acct: or https:)
            domain: Optional domain override

        Returns:
            WebFinger response data
        """
        try:
            if not domain:
                if resource.startswith("acct:"):
                    domain = resource.split("@")[1]
                else:
                    domain = urlparse(resource).netloc

            url = urljoin(domain, "/.well-known/webfinger?resource={resource}")

            response = await self.requests.get(url, headers=self.headers)
            if response.status_code != 200:
                return None

            return cast(dict[str, Any], response.json())

        except Exception as e:
            self.logger.error(f"WebFinger lookup failed for {resource}: {e}")
            return None

    def close(self) -> None:
        """Clean up resources."""
        pass



from http import HTTPStatus
from typing import Annotated, override

from fastapi import HTTPException, Security
from fastapi.responses import JSONResponse

from phederation.api.routes.base import BaseRoute, admin_route
from phederation.federation.instance import InstanceCollectionType
from phederation.models.activities import APActivity, ErrorMessageActivityValidation
from phederation.models.actors import ErrorMessageActorUnauthorized
from phederation.models.collections import APOrderedCollection, APOrderedCollectionPage
from phederation.security.authentication import UserInfo
from phederation.utils import ActivityType
from phederation.utils.base import ObjectId, UrlType
from phederation.utils.exceptions import ActivityPubException, DeliveryError


class AdminInstanceRoute(BaseRoute):

    SUPPORTED_TYPES: list[ActivityType] = [ActivityType.MAINTENANCE, ActivityType.DELETE]

    @override
    def setup(self):

        @admin_route(self)
        @self.router.get(f"/{UrlType.Instance.value}/{{collection}}", tags=["admin"])
        async def get_instance_collections(  # pyright: ignore[reportUnusedFunction]
            collection: InstanceCollectionType, page: int = 1
        ) -> APOrderedCollection | APOrderedCollectionPage:
            """Get special instance collections, like federated/blocked instances.

            Only possible in admin mode.
            """
            instance_actor_id = self.server.instance.actor_id
            collection_id: ObjectId | None = None
            if collection == InstanceCollectionType.Federated:
                collection_id = self.server.instance.federated_instances
            elif collection == InstanceCollectionType.Blocked:
                collection_id = self.server.instance.blocked_instances
            elif collection == InstanceCollectionType.BannedActors:
                collection_id = self.server.instance.banned_actors
            elif collection == InstanceCollectionType.FederatedSharedInboxes:
                collection_id = self.server.instance.federated_instance_actors
            elif collection == InstanceCollectionType.MaintenanceCommands:
                collection_id = self.server.instance.maintenance_commands
            elif collection == InstanceCollectionType.ActivitiesIncoming:
                collection_id = self.server.instance.activities_incoming
            elif collection == InstanceCollectionType.ActivitiesOutgoing:
                collection_id = self.server.instance.activities_outgoing

            if collection_id:
                # TODO: if the second page or more are accessed then the collection ID is not the admin id anymore - fix
                collection_obj = await self.server.handle_pagination(
                    collection_id=collection_id, actor_id=instance_actor_id, actor_for_authorization=instance_actor_id, page=page
                )
                return collection_obj
            else:
                raise HTTPException(status_code=HTTPStatus.BAD_REQUEST, detail=f"Collection {collection} does not exist")

        @self.router.post(
            "/maintenance",
            tags=["admin"],
            responses={
                HTTPStatus.BAD_REQUEST: {
                    "model": ErrorMessageActivityValidation,
                    "description": ErrorMessageActivityValidation().description,
                },
                HTTPStatus.UNAUTHORIZED: {
                    "model": ErrorMessageActorUnauthorized,
                    "description": ErrorMessageActorUnauthorized().description,
                },
            },
        )
        async def post_admin_maintenance(  # pyright: ignore[reportUnusedFunction]
            activity: APActivity,
            user_info: Annotated[UserInfo, Security(self.middleware.authorization, scopes=["admin"])],
        ):
            """
            Handle admin maintenance requests (client-to-server).

            The POST is secured by authentication ("admin" scope), the admin user need to authenticate using Bearer token (Oauth2).
            """
            # set actor who this was posted to as the owner of the activity.
            activity.actor = user_info.actor_id

            if not ActivityType(activity.type) in AdminInstanceRoute.SUPPORTED_TYPES:
                return JSONResponse(
                    content={
                        "Status": "Error",
                        "description": f"Activity is not in {AdminInstanceRoute.SUPPORTED_TYPES}, cannot handle in /maintenance route",
                    },
                    status_code=HTTPStatus.BAD_REQUEST,
                    media_type="application/activity+json",
                )

            try:
                activity = await self.server.process_outbox_activity(activity=activity, is_maintenance=True)
                if not activity.id:
                    raise DeliveryError("Activity id is not set")
            except ActivityPubException as e:
                self.logger.debug(f"Handle outbox failed with {type(e).__name__} {e.message}. Failing silently (HttpStatus.CONTINUE).")
                e.status_code = HTTPStatus.CONTINUE
                raise e

            # Servers MUST return a 201 Created HTTP code, and unless the activity is transient, MUST include the new id in the Location header.
            # https://www.w3.org/TR/activitypub/#client-to-server-interactions
            return JSONResponse(
                content={"Status": "OK"},
                status_code=HTTPStatus.CREATED,
                media_type="application/activity+json",
                headers={"Location": activity.id},
            )



"""The actions module.

Implements workflows for asynchronous processing in temporalio.
"""

from .worker import main_worker
from .workflows import (
    ActivityDeliveryWorkflow,
    PrepareWorkflowParams,
    UpdateNodeInfoWorkflow,
)
from .workflows import DeliveryWorkerTask, KeyRotationTask, UpdateNodeInfoTask

__all__ = [
    "KeyRotationTask",
    "DeliveryWorkerTask",
    "UpdateNodeInfoTask",
    "ActivityDeliveryWorkflow",
    "PrepareWorkflowParams",
    "UpdateNodeInfoWorkflow",
    "main_worker",
]



"""
Activity handlers module for processing ActivityPub activities.
"""

from .accept import AcceptHandler
from .addremove import AddRemoveHandler
from .announce import AnnounceHandler
from .base import ActivityHandler
from .block import BlockHandler
from .create import CreateHandler
from .delete import DeleteHandler
from .follow import FollowHandler
from .like import LikeHandler
from .media import MediaHandler
from .reject import RejectHandler
from .undo import UndoHandler
from .update import UpdateHandler

__all__ = [
    "AddRemoveHandler",
    "ActivityHandler",
    "CreateHandler",
    "FollowHandler",
    "LikeHandler",
    "DeleteHandler",
    "AnnounceHandler",
    "UpdateHandler",
    "UndoHandler",
    "AcceptHandler",
    "RejectHandler",
    "BlockHandler",
    "MediaHandler",
]



"""
Enhanced Create activity handler.
"""

import io
import json
from typing import Any, cast, override

from fastapi.datastructures import UploadFile

from phederation.handlers.create import CreateHandler
from phederation.models import dereference, object_from_type
from phederation.models.activities import APActivity, APMigrate
from phederation.models.actors import APActor
from phederation.models.objects import (
    APDocument,
    APObject,
    DocumentType,
    dereference_or_raise,
)
from phederation.utils import UrlType
from phederation.utils.base import AccessType
from phederation.utils.exceptions import HandlerError, ValidationError
from phederation.utils.serialization import ActivityPubBase, to_json_string

from .base import ActivityHandler


class MigrationHandler(ActivityHandler):
    """Migration activity handler."""

    SUPPORTED_TYPES: dict[str, type[APDocument]] = {
        DocumentType.DOCUMENT.value: APDocument,
    }

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """
        Migration validation.

        Validates:
        - Activity structure
        - Object type support
        - Rate limits
        """
        activity_resolved = object_from_type(activity)
        if not isinstance(activity_resolved, APMigrate):
            raise ValidationError(f"APActivity is not APMigrate (is type {activity.type})")
        activity = activity_resolved

        # Check actor permissions
        actor_id = dereference(activity.actor, key="id")
        if not actor_id:
            raise ValidationError("APMigrate.actor.id is None")

        actor = await self.resolver.resolve_actor(actor_id=actor_id)
        if not actor:
            raise ValidationError("Actor not found")

        migration_document = activity.object

        # if there is a migration document attached, validate it and store it locally
        if migration_document:
            if isinstance(migration_document, APObject):
                migration_document = migration_document.serialize()
            if not isinstance(migration_document, dict):
                self.logger.warning(f"Object in activity was not an APObject (instead: {type(migration_document)}): activity {activity.serialize()}")
                raise ValidationError(f"Object in MIGRATE did not have proper information")
            obj_type = cast(str, migration_document.get("type", ""))

            if not obj_type or obj_type not in self.SUPPORTED_TYPES:
                raise ValidationError(f"Unsupported object type: {obj_type}")

            # Validate object model
            model_class = self.SUPPORTED_TYPES[obj_type]
            obj = model_class.model_validate(migration_document)

            if obj.attributed_to and obj.attributed_to != actor_id:
                raise ValidationError(f"Object in Migrate is attributed to the wrong actor: attributed_to={obj.attributed_to}, actor_id={actor.id}")
            obj.attributed_to = actor_id

            # Validate file hash
            await CreateHandler.validate_content_hash(self, obj)

            # Visibility of the object cannot be less secure than activity
            activity_access = AccessType.from_visibility(activity.visibility)
            object_access = AccessType.from_visibility(obj.visibility)
            maximum_security_visibility = AccessType.maximum_security(object_access, activity_access).value

            # Set the visibility of the activity to be at most the one from the object
            obj.visibility = maximum_security_visibility
            activity.visibility = maximum_security_visibility

            # Important to do this not too late in the handle_outbox method
            activity.object = await CreateHandler.store_object_locally(handler=self, obj=obj, actor_id=actor_id)

        # Set the accepted field to None, so that it must be set explicitly by an APAccept
        activity.accepted_at = None

        return activity

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """
        Migration activity processing in the instance the actor is *leaving*.

        Handles:
        - Assembling the data of the migrating actor
        - Informing the new instance about the user data being ready for migration
        """
        if not isinstance(activity, APMigrate):
            raise HandlerError(f"Activity is not of type Migrate, is {activity.type}")

        activity = await self._assemble_user_data(activity)
        activity = await self._inform_new_instance(activity)

        return activity

    async def _assemble_user_data(self, activity: APMigrate):
        actor_id = dereference_or_raise(activity, "actor")
        activity_id = dereference_or_raise(activity, "id")

        actor = await self.resolver.try_resolve_from_storage(url=actor_id, url_type=UrlType.Users)
        if not isinstance(actor, APActor):
            raise ValidationError("APMigrate.actor could not be resolved from storage")
        document = await self.create_user_data(actor)
        document.visibility = activity.visibility

        activity.object = document
        activity.id = await self.storage.activity.upsert(id=activity_id, data=activity)
        return activity

    async def _inform_new_instance(self, activity: APMigrate):
        """Inform the instance the actor is migrating to about the migration data being ready.

        Args:
            activity (APMigrate): The migration activity.

        Raises:
            HandlerError: If the activity is not properly set up.

        Returns:
            APMigrate: The activity, without any recipients.
        """
        object_id = dereference(activity, "object")
        if not object_id:
            raise HandlerError(f"APMigrate activity {activity.id} does not have a proper object with id, cannot migrate")
        if not activity.actor_to:
            self.logger.info(
                "Id of the actor on the new instance is not set in the APMigrate activity, will not inform any new instance about the migration data"
            )
            return activity

        # do not send to anybody except the new actor
        activity.clear_recipients()
        result = await self.delivery.deliver_activity(activity=activity, recipients=[activity.actor_to])
        if result.failed:
            self.logger.info(f"Failed to deliver message to the new actor {activity.actor_to}. Message: {result.error_message}")
        return activity

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        """
        Migration activity processing in the instance the actor is *migrating to*.

        This only validates the document proof created by the old instance.
        The actor on this instance must send an APAccept to this instance with the object
        begin the APMigrate activity id. This then triggers the migration.
        """
        self.logger.info(f"Received MIGRATE in inbox (activity id: {activity.id})", extra={"activity_id": activity.id})

        # Always validate content proof
        self.logger.debug(f"Validate file proof in inbox", extra={"activity_id": activity.id})
        obj = await self.resolver.resolve_object(activity.object)
        await CreateHandler.validate_proof(self, obj)

        return None

    async def create_user_data(self, actor: APActor) -> APDocument:
        self.logger.info("Received Migrate activity. Creating user data")
        actor_id = actor.id
        if not actor_id:
            raise HandlerError("APMigrate.actor is None, cannot create user data file")
        actor_serialized = to_json_string(actor.serialize())

        document = APDocument()

        document.media_type = "application/json"
        document.name = actor_id

        # create object in storage so that it gets and id
        document = await CreateHandler.store_object_locally(handler=self, obj=document, actor_id=actor_id)
        if not document.id:
            raise HandlerError("Descriptor document for Actor does not have a proper id, cannot store actor as file")
        if not isinstance(document, APDocument):
            raise HandlerError("Descriptor document for Actor is not a proper APDocument, cannot store actor as file")

        filename = actor_id
        inmemory_file = UploadFile(filename=filename, file=io.BytesIO(actor_serialized.encode()))
        file_id = await self.storage.save(document_id=document.id, content=inmemory_file)
        document.url = [file_id]

        # get metadata that was saved with the file
        file_metadata = await self.storage.metadata(file_id=file_id)
        document.hash = file_metadata.hash_sha256

        if not document.hash:
            raise HandlerError("Not able to create SHA256 hash for file content")

        # create a proof of the file content, sign it with the actor key
        actor_current_key = await self.key_manager.get_key_id_from_actor(actor_id)
        document.proof = await self.key_manager.proof_string(string_to_sign=document.hash, private_key_id=actor_current_key)
        document.proof.description = f"SHA256 hash of the file content, signed by actor {actor_id}"

        # update the object with the file info
        _ = await self.storage.object.upsert(document.id, data=document)

        return document

    @staticmethod
    async def accept_migrate_on_new_instance(handler: ActivityHandler, activity: APMigrate):
        """Accept the migration event on the new instance.
        This will create an actor if they are not already created, and sets their memorial status to False.

        Args:
            activity (APMigrate): The APMigrate activity.
        """
        if not activity.actor_to:
            raise HandlerError(f"APMigrate.actor_to must not be None (activity {activity.id})")

        file_document = await handler.resolver.resolve_object(activity.object)
        file_urls = file_document.url
        if not file_urls:
            raise HandlerError("file_document.url should contain the url to the migration file (activity {activity.id})")

        # download the data from the old instance
        migration_objects: list[ActivityPubBase] = []

        # TODO: what is a good way to pass a user token for the old instance? Right now the user data must be publicly accessible...
        token = None
        for file_url in file_urls:
            resolved_file = b"".join([line async for line in await handler.resolver.resolve_media(file_id=file_url, access_token=token)])

            try:
                object_dict = cast(dict[str, Any], json.loads(resolved_file.decode()))
                migration_objects.append(object_from_type(object_dict))
            except json.JSONDecodeError:
                raise HandlerError(f"Unable to download user data from old instance during APMigrate (JSONDecodeError, activity {activity.id})")

        actor_object = None
        # account_object = None

        for obj in migration_objects:
            if isinstance(obj, APActor):
                actor_object = obj
            # if isinstance(obj, APAccount):
            #    account_object = obj

        if not actor_object:
            raise HandlerError(f"Files in migration data did not contain APActor data ({activity.id})")
    
        # TODO: instead of create actor update the actor_to actor here!
        _ = await handler.actor_manager.update_actor(update=actor_object, actor_id=activity.actor_to)
        # _ = self.actor_manager.create_account(activity=APCreate(object=account_object, actor=activity.actor_to))

    @staticmethod
    async def accept_migrate_on_previous_instance(handler: ActivityHandler, activity: APMigrate):
        """Handle migration on the previous instance of the migrating actor.
        This will disable the actor on the instance by setting their memorial=True and discoverable=False.

        Args:
            activity (APMigrate): The migration activity.
        """
        actor_id = dereference_or_raise(activity, "actor")
        if not (await handler.actor_manager.deactivate_actor(actor_id)):
            handler.logger.info(f"Unable to deactivate actor {actor_id} on the previous instance during migration")
        else:
            handler.logger.info(f"Deactivated actor {actor_id} on the previous instance during migration")



"""
maintenance.py
Defines additional objects that can be used for maintenance of Phederation instances.
These objects are *not* based on the ActivityStreams vocabulary, they extend the specification.
"""

from abc import ABC
import enum
from typing import override

from pydantic import Field

from phederation.models.activities import APActivity
from phederation.models.objects import APObject
from phederation.utils import ActivityType
from phederation.utils.base import APDateTime, ObjectId


class MaintenanceCommandType(enum.Enum):
    """A subset of types that concern instance maintenance."""

    MAINTENANCE_MODE = "MaintenanceMode"
    MAINTENANCE_MODE_CANCEL = "MaintenanceModeCancel"
    ASSEMBLE_USER_DATA = "AssembleUserData"

    @override
    def __str__(self):
        return str(self.value)


class APMaintenance(APActivity):
    """
    Represents a maintenance activity.
    This extends the ActivityStreams vocabulary to include maintenance information.
    Sending this activity (by an admin, to an instance in admin_mode) executes maintenance commands.
    """

    type: str = ActivityType.MAINTENANCE.value


class APMaintenanceCommand(ABC, APObject):
    """
    Represents an abstract maintenance command.
    """


class APMaintenanceMode(APMaintenanceCommand):
    """
    Represents a maintenance mode command.
    Sending this activity (by an admin, to an instance in admin_mode) will set the related main instance into "maintenance mode".
    The start_time and end_time can be set to start and end the maintenance mode. If they are not set, the mode is started immediately and does not end automatically (must be ended by sending APDelete for the APMaintenance object).
    """

    type: str = MaintenanceCommandType.MAINTENANCE_MODE.value
    start_time: APDateTime | None = Field(default=None, description="Start datetime of the maintenance mode")
    end_time: APDateTime | None = Field(default=None, description="End datetime of the maintenance mode")


class APMaintenanceModeCancel(APMaintenanceCommand):
    """
    Cancels a specific maintenance mode command that was issued earlier.
    This is different from a maintenance mode running over its end_time; the cancel command can stop maintenance mode that is currently running.
    """

    type: str = MaintenanceCommandType.MAINTENANCE_MODE_CANCEL.value
    command_to_cancel: ObjectId = Field(..., description="The id of the command object that should be cancelled")



"""Accept activity handler."""

from typing import override

from phederation.models import dereference
from phederation.models.activities import APActivity
from phederation.models.collections import ValidCollection
from phederation.models.objects import dereference_or_raise
from phederation.models.proofs import DataIntegrityProof
from phederation.utils import NodeInfo, UrlType
from phederation.utils.base import (
    AccessType,
    ActivityPubBaseWithId,
    ActivityType,
    collection_id_from_name,
    urljoin,
)
from phederation.utils.exceptions import AuthorizationError, HandlerError, ValidationError

from .base import ActivityHandler


class AddRemoveHandler(ActivityHandler):
    """Handles Add and Remove activities."""

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """Validate Add/Remove activity."""
        if not (activity.type in [ActivityType.ADD.value, ActivityType.REMOVE.value]):
            raise ValidationError("Invalid activity type")

        actor_id = dereference(activity, "actor")
        if not actor_id:
            raise ValidationError("Add/Remove activity must have an actor set")

        object_id = dereference_or_raise(activity, "object")

        # TODO: also make it possible to add remote objects to lists
        object_url_type = UrlType.from_local_url(object_id)
        object_to_handle = await self.resolver.try_resolve_from_storage(object_id, url_type=object_url_type)

        if not object_to_handle:
            raise ValidationError("Add/Remove activity must have an existing object to add/remove")

        # validate access
        obj_access = AccessType.from_visibility(object_to_handle.visibility)
        obj_attribution = dereference(object_to_handle, "attributedTo")
        if not obj_access.is_accessible(AccessType.PUBLIC) and (not obj_attribution or obj_attribution != actor_id):
            raise AuthorizationError(f"Not authorized to Add/Remove this object (obj_access={obj_access}, attributed_to={obj_attribution})")
        
        if isinstance(object_to_handle, DataIntegrityProof):
            raise ValidationError("Cannot add/remove DataIntegrityProof to a list")
        if isinstance(object_to_handle, NodeInfo):
            raise ValidationError("Cannot add/remove NodeInfo to a list")
        
        activity.object = object_to_handle

        if activity.type == ActivityType.REMOVE.value:
            collection_id = dereference(activity, "origin")
            if not collection_id:
                raise ValidationError("Remove activity must have an origin set where to remove the object")
        elif activity.type == ActivityType.ADD.value:
            collection_id = dereference(activity, "target")
            if not collection_id:
                raise ValidationError("Add activity must have a target set where to add the object")
        else:
            collection_id = None
        if not collection_id:
            raise ValidationError("Add/Remove activity does not have target/origin id set")

        # validate that the collection id is part of the actor collections
        if not collection_id.startswith(urljoin(actor_id, ValidCollection.Collections.value)):
            raise ValidationError(f"Collection id {collection_id} is not a collection of actor {actor_id}, cannot Add/Remove")

        return activity

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """Handle Add and Remove activities."""
        actor_id = dereference_or_raise(activity, "actor")
        actor_collections = collection_id_from_name(id=actor_id, name=ValidCollection.Collections.value)
        object_to_handle = activity.object
        if not isinstance(object_to_handle, ActivityPubBaseWithId):
            raise HandlerError(f"Activity.object to Add/Remove has wrong type: {type(object_to_handle).__name__}")
        
        object_id = dereference_or_raise(object_to_handle, "id")

        # set the access level to the maximum security between activity and object
        access_of_activity = AccessType.from_visibility(activity.visibility)
        access_of_object = AccessType.from_visibility(object_to_handle.visibility)
        access = AccessType.maximum_security(access_of_activity, access_of_object)

        if activity.type == ActivityType.ADD.value:
            collection_id = dereference_or_raise(activity, "target")
            _ = await self.collections.add_to_collection(collection_id=collection_id, items=object_id, access=access, add_only_once=False)
            _ = await self.collections.add_to_collection(collection_id=actor_collections, items=collection_id, access=access, add_only_once=True)
            self.logger.info(f"Added {object_id} to actor collection {collection_id}")
        elif activity.type == ActivityType.REMOVE.value:
            collection_id = dereference_or_raise(activity, "origin")
            _ = await self.collections.remove_from_collection(collection_id=collection_id, items=object_id, access=access)
            self.logger.info(f"Removed {object_id} from actor collection {collection_id} once")
            if not (await self.collections.collection_exists(collection_id=collection_id)):
                await self.collections.remove_from_collection(collection_id=actor_collections, items=collection_id, access=access)
                self.logger.info(f"Collection is now empty, also removing {collection_id} from actor collections")

        # TODO: notify followers?
        return activity

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        """Process inbox for APAdd and APRemove."""
        # Nothing to do here? This would arise if followers are notified.
        object_id = dereference_or_raise(activity, "object")
        self.logger.info(f"Received {activity.type} activity for object {object_id} in inbox, nothing to do here")

        return activity.id



# logging.py

"""
Uses the [python `logging` facility](https://docs.python.org/3/library/logging.html) to set up loggers for all classes.
The configuration file is stored separately in `settings/logging.yaml` and defines where the log output is being written to.
"""

import logging
from typing import override


class InstanceFilter(logging.Filter):
    def __init__(self, error_only: list[str], logging_prefix: str | None = None, name: str = "") -> None:
        super().__init__(name)
        self.logging_prefix: str | None = logging_prefix
        self.error_only: list[str] = error_only

    @override
    def filter(self, record: logging.LogRecord):
        for check in self.error_only:
            if check == record.name and record.levelno < logging.ERROR:
                return False
        return (record.name.split("|")[0].startswith(self.logging_prefix)) if self.logging_prefix else True


def configure_logger(name: str, prefix: str = "phederation") -> logging.Logger:
    """
    Get a logger with the specified name.

    Args:
        name (str): The name of the logger, typically __name__ of the module.
        prefix (str): the prefix to use for the logger name. Default "phederation"

    Returns:
        logging.Logger: A logger instance.
    """
    return logging.getLogger(f"{prefix}|{name}")



from http import HTTPStatus
from typing import Annotated, override

from fastapi import HTTPException, Security

from phederation.api.routes.base import BaseRoute
from phederation.models import (
    ErrorMessageActivityNotFound,
    ErrorMessageActivityRestricted,
)
from phederation.models.activities import APActivity
from phederation.security.authentication import UserInfo
from phederation.utils.base import UrlType, assemble_id_url


class ActivityRoute(BaseRoute):

    @override
    def setup(self):

        @self.router.get(
            "/activities/{activity_id}",
            tags=["activities"],
            status_code=HTTPStatus.OK,
            response_model_exclude_none=True,
            responses={
                HTTPStatus.NOT_FOUND: {
                    "model": ErrorMessageActivityNotFound,
                    "description": ErrorMessageActivityNotFound().description,
                },
                HTTPStatus.BAD_REQUEST: {
                    "model": ErrorMessageActivityRestricted,
                    "description": ErrorMessageActivityRestricted().description,
                },
            },
        )
        async def get_activity(  # pyright: ignore[reportUnusedFunction]
            activity_id: str, user_info: Annotated[UserInfo | None, Security(self.middleware.authentication, scopes=["user"])] = None
        ) -> APActivity:
            """Return activity with given id, if accessible."""
            activity_id = assemble_id_url(
                base_url=self.api.settings.domain.base_url,
                primary=activity_id,
                type=UrlType.Activities,
            )
            activity: APActivity | None = await self.server.storage.activity.read(id=activity_id)
            if not activity:
                raise HTTPException(
                    status_code=HTTPStatus.NOT_FOUND,
                    detail=ErrorMessageActivityNotFound().message,
                )

            actor_requesting_access = user_info.actor_id if user_info else None
            activity_with_access = await self.api.server.resolver.resolve_access_type(
                object=activity, actor_requesting_access=actor_requesting_access
            )
            if not activity_with_access or not isinstance(activity_with_access, APActivity):
                self.logger.warning(f"Activity visibility restricted, user: {user_info} cannot access it")
                raise HTTPException(
                    status_code=HTTPStatus.BAD_REQUEST,
                    detail=ErrorMessageActivityRestricted().message,
                )
            return activity_with_access



from logging import Logger
from typing import Any
from urllib.parse import urlparse

from fastapi import FastAPI, Request, Response
from starlette.background import BackgroundTask

from phederation.storage.base import StorageBackend
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings


class Statistics:
    """Statistics class.
    Can be used for logging url requests and corresponding responses from the API, for debugging or for web statistics.
    CAREFUL: This can be used to store sensitive information, like authentication tokens and hidden messages from users, so use carefully.
    The StatisticsSettings can be used to disable certain routes, like '/auth' and '/keys', to mitigate this.
    """

    def __init__(self, settings: PhedSettings, app: FastAPI, storage: StorageBackend):
        self.app: FastAPI = app
        self.settings: PhedSettings = settings
        self.logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix)

    def log_statistics(
        self,
        extra: dict[str, Any],
        req_headers: dict[str, str | None] | None = None,
        res_headers: dict[str, str | None] | None = None,
    ):
        path = urlparse(str(extra["req"]["url"])).path  # pyright: ignore[reportAny]
        self.logger.info(f"Statistics: logging '{path}', req_headers={req_headers}, res_headers={res_headers}, extra: {extra}")

    async def initialize(self):
        if self.settings.statistics.enabled:
            self.logger.info("Initializing Statistics module.")

    async def log(self, request: Request, response: Response):
        """
        Logs statistics about the given request and response. Called in middleware.
        Inspired by https://stackoverflow.com/questions/69670125/how-to-log-raw-http-request-response-in-fastapi/73464007#73464007
        """
        if not self.settings.statistics.enabled or not self.settings.statistics.log_middleware:
            return response
        else:
            path = urlparse(str(request.url)).path

            req_headers_obj = request.headers
            req_headers = {key.lower(): req_headers_obj.get(key) for key in req_headers_obj.keys()}
            res_headers_obj = response.headers
            res_headers = {key.lower(): res_headers_obj.get(key) for key in res_headers_obj.keys()}

            # avoid logging authorization tokens
            if "authorization" in req_headers.keys():
                req_headers["authorization"] = "******"
            if "authorization" in res_headers.keys():
                res_headers["authorization"] = "******"

            if not self.settings.statistics.disable_routes or not any([path.startswith(route) for route in self.settings.statistics.disable_routes]):
                extra = {
                    "req": {"method": request.method, "url": str(request.url)},
                    "res": {
                        "status_code": response.status_code,
                        "media_type": response.media_type,
                    },
                }
                task = BackgroundTask(
                    self.log_statistics,
                    req_headers=req_headers,
                    res_headers=res_headers,
                    extra=extra
                )
            else:
                task = None
            response.background = task
            return response



"""Security module.

Implements security measures for HTTP signatures and temporalio encryption.
"""

from .authentication import Authentication
from .authorization import Authorization, PermissionType
from .http_signatures import HTTPSignatureVerifier
from .key_management import KeyManager
from .worker_encryption import EncryptionCodec

__all__ = ["KeyManager", "HTTPSignatureVerifier", "EncryptionCodec", "Authentication", "PermissionType", "Authorization"]
